import struct
import os.path as path
import numpy as np
from logging import getLogger
import warnings
from .. import time
from .base import _abspath


def _reduce_by_average(data, ky0, ky1):
    # Average two arrays together, if they both exist.
    if ky1 in data:
        tmp = data.pop(ky1)
        if ky0 in data:
            data[ky0] += tmp
            data[ky0] = data[ky0] / 2
        else:
            data[ky0] = tmp


def _reduce_by_average_angle(data, ky0, ky1, degrees=True):
    # Average two arrays of angles together, if they both exist.
    if degrees:
        rad_fact = np.pi / 180
    else:
        rad_fact = 1
    if ky1 in data:
        if ky0 in data:
            data[ky0] = np.angle(
                np.exp(1j * data.pop(ky0) * rad_fact) +
                np.exp(1j * data.pop(ky1) * rad_fact)) / rad_fact
        else:
            data[ky0] = data.pop(ky1)


# This is the data-type of the index file.
# This must match what is written-out by the create_index function.
_index_version = 1
_hdr = struct.Struct('<BBBBhhh')
_index_dtype = {
    None:
    np.dtype([('ens', np.uint64),
              ('pos', np.uint64),
              ('ID', np.uint16),
              ('config', np.uint16),
              ('beams_cy', np.uint16),
              ('_blank', np.uint16),
              ('year', np.uint8),
              ('month', np.uint8),
              ('day', np.uint8),
              ('hour', np.uint8),
              ('minute', np.uint8),
              ('second', np.uint8),
              ('usec100', np.uint16),
              ]),
    1:
    np.dtype([('ens', np.uint64),
              ('hw_ens', np.uint32),
              ('pos', np.uint64),
              ('ID', np.uint16),
              ('config', np.uint16),
              ('beams_cy', np.uint16),
              ('_blank', np.uint16),
              ('year', np.uint8),
              ('month', np.uint8),
              ('day', np.uint8),
              ('hour', np.uint8),
              ('minute', np.uint8),
              ('second', np.uint8),
              ('usec100', np.uint16),
              ('d_ver', np.uint8),
              ])
}


def _calc_time(year, month, day, hour, minute, second, usec, zero_is_bad=True):
    dt = np.zeros(year.shape, dtype='float')
    for idx, (y, mo, d, h, mi, s, u) in enumerate(
            zip(year, month, day,
                hour, minute, second, usec)):
        if (zero_is_bad and mo == 0 and d == 0 and
                h == 0 and mi == 0 and
                s == 0 and u == 0):
            continue
        try:
            # Note that month is zero-based, seconds since Jan 1 1970
            dt[idx] = time.date2epoch(
                time.datetime(y, mo + 1, d, h, mi, s, u))[0]
        except ValueError:
            # One of the time values is out-of-range (e.g., mi > 60)
            # This probably indicates a corrupted byte, so we just insert None.
            dt[idx] = None
    # None -> NaN in this step
    return dt


def _create_index(infile, outfile, N_ens, debug):
    logging = getLogger()
    print("Indexing {}...".format(infile), end='')
    fin = open(_abspath(infile), 'rb')
    fout = open(_abspath(outfile), 'wb')
    fout.write(b'Index Ver:')
    fout.write(struct.pack('<H', _index_version))
    ens = dict.fromkeys([21, 23, 24, 26, 28], 0)
    N = dict.fromkeys([21, 23, 24, 26, 28], 0)
    config = 0
    last_ens = dict.fromkeys([21, 23, 24, 26, 28], -1)
    seek_2ens = {21: 40, 24: 40, 26: 40,
                 23: 42, 28: 40}  # 23 starts from "42"
    while N[21] < N_ens:  # Will fail if velocity ping isn't saved first
        pos = fin.tell()
        try:
            dat = _hdr.unpack(fin.read(_hdr.size))
        except:
            break
        if dat[2] in [21, 23, 24, 26, 28]:  # vel, bt, vel_b5, alt_raw, echo
            idk = dat[2]
            d_ver, d_off, config = struct.unpack('<BBH', fin.read(4))
            fin.seek(4, 1)
            yr, mo, dy, h, m, s, u = struct.unpack('6BH', fin.read(8))
            fin.seek(14, 1)
            beams_cy = struct.unpack('<H', fin.read(2))[0]
            fin.seek(seek_2ens[dat[2]], 1)
            ens[idk] = struct.unpack('<I', fin.read(4))[0]

            if ens[idk] == 1 and last_ens[idk] > 0:
                # Covers all id keys saved in "burst mode"
                ens[idk] = last_ens[idk]+1

            if last_ens[idk] > 0 and last_ens[idk] != ens[idk]:
                N[idk] += 1

            fout.write(struct.pack('<QIQ4H6BHB', N[idk], ens[idk], pos, idk,
                                   config, beams_cy, 0,
                                   yr, mo + 1, dy, h, m, s, u, d_ver))
            fin.seek(dat[4] - (36 + seek_2ens[idk]), 1)
            last_ens[idk] = ens[idk]

            if debug and N[idk] < 5:
                # hex: [18, 15, 1C, 17] = [vel_b5, vel, echo, bt]
                logging.info('%10d: %02X, %d, %02X, %d, %d, %d, %d\n' %
                             (pos, dat[0], dat[1], dat[2], dat[4],
                             N[idk], ens[idk], last_ens[idk]))
        else:
            fin.seek(dat[4], 1)
    fin.close()
    fout.close()
    print(" Done.")


def _check_index(idx, infile, fix_hw_ens=False):
    uid = np.unique(idx['ID'])
    if fix_hw_ens:
        hwe = idx['hw_ens']
    else:
        hwe = idx['hw_ens'].copy()
    period = hwe.max()
    ens = idx['ens']
    N_id = len(uid)
    FLAG = False
    # This loop fixes 'skips' inside the file
    for id in uid:
        # These are the indices for this ID
        inds = np.nonzero(idx['ID'] == id)[0]

        # These are bad steps in the indices for this ID
        ibad = np.nonzero(np.diff(inds) > N_id)[0]
        for ib in ibad:
            FLAG = True
            # The ping number reported here may not be quite right if
            # the ensemble count is wrong.
            warnings.warn("Skipped ping (ID: {}) in file {} at ensemble {}."
                          .format(id, infile, idx['ens'][inds[ib + 1] - 1]))
            hwe[inds[(ib + 1):]] += 1
            ens[inds[(ib + 1):]] += 1

    # This block fixes skips that originate from before this file.
    delta = max(hwe[:N_id]) - hwe[:N_id]
    for d, id in zip(delta, idx['ID'][:N_id]):
        if d != 0:
            FLAG = True
            hwe[id == idx['ID']] += d
            ens[id == idx['ID']] += d

    if np.any(np.diff(ens) > 1) and FLAG:
        idx['ens'] = np.unwrap(hwe.astype(np.int64), period=period) - hwe[0]


def _boolarray_firstensemble_ping(index):
    """Return a boolean of the index that indicates only the first ping in 
    each ensemble.
    """
    dens = np.ones(index['ens'].shape, dtype='bool')
    dens[1:] = np.diff(index['ens']) != 0
    return dens


def get_index(infile, reload=False, debug=False):
    """This function reads ad2cp.index files

    Parameters
    ----------
    infile: str
      Path and filename of ad2cp datafile, not including ".index"
    reload: bool
      If true, ignore existing .index file and create a new one
    debug: bool
      If true, run code in debug mode

    Returns
    -------
    out: tuple
      Tuple containing info held within index file

    """
    index_file = infile + '.index'
    if not path.isfile(index_file) or reload:
        _create_index(infile, index_file, 2 ** 32, debug)
    f = open(_abspath(index_file), 'rb')
    file_head = f.read(12)
    if file_head[:10] == b'Index Ver:':
        index_ver = struct.unpack('<H', file_head[10:])[0]
    else:
        # This is pre-versioning the index files
        index_ver = None
        f.seek(0, 0)
    out = np.fromfile(f, dtype=_index_dtype[index_ver])
    f.close()
    _check_index(out, infile)
    return out


def crop_ensembles(infile, outfile, range):
    """This function is for cropping certain pings out of an AD2CP
    file to create a new AD2CP file. It properly grabs the header from
    infile.

    The range is the `ensemble/ping` counter as defined in the first column
    of the INDEX.

    Parameters
    ----------
    infile: str
      Path of ad2cp filename (with .ad2cp file extension)
    outfile: str
      Path for new, cropped ad2cp file (with .ad2cp file extension)
    range: list
      2 element list of start and end ensemble (or time index)

    """
    idx = get_index(infile)
    with open(_abspath(infile), 'rb') as fin:
        with open(_abspath(outfile), 'wb') as fout:
            fout.write(fin.read(idx['pos'][0]))
            i0 = np.nonzero(idx['ens'] == range[0])[0][0]
            ie = np.nonzero(idx['ens'] == range[1])[0][0]
            pos = idx['pos'][i0]
            nbyte = idx['pos'][ie] - pos
            fin.seek(pos, 0)
            fout.write(fin.read(nbyte))


class _BitIndexer():
    def __init__(self, data):
        self.data = data

    @property
    def _data_is_array(self, ):
        return isinstance(self.data, np.ndarray)

    @property
    def nbits(self, ):
        if self._data_is_array:
            return self.data.dtype.itemsize * 8
        else:
            raise ValueError("You must specify the end-range "
                             "for non-ndarray input data.")

    def _get_out_type(self, mask):
        # The mask indicates how big this item is.
        if not self._data_is_array:
            return None
        if mask < 2:
            return bool
        if mask < 2 ** 8:
            return np.uint8
        elif mask < 2 ** 16:
            return np.uint16
        elif mask < 2 ** 32:
            return np.uint32
        else:
            return np.uint64

    def __getitem__(self, slc):
        if isinstance(slc, int):
            slc = slice(slc, slc + 1)
        if slc.step not in [1, None]:
            raise ValueError("Slice syntax for `_getbits` does "
                             "not support steps")
        start = slc.start
        stop = slc.stop
        if start is None:
            start = 0
        if stop is None:
            stop = self.nbits
        mask = 2 ** (stop - start) - 1
        out = (self.data >> start) & mask
        ot = self._get_out_type(mask)
        if ot is not None:
            out = out.astype(ot)
        return out


def _getbit(val, n):
    return bool((val >> n) & 1)


def _headconfig_int2dict(val, mode='burst'):
    """Convert the burst Configuration bit-mask to a dict of bools.

    mode: {'burst', 'bt'}
       For 'burst' configs, or 'bottom-track' configs.
    """
    if mode == 'burst':
        return dict(
            press_valid=_getbit(val, 0),
            temp_valid=_getbit(val, 1),
            compass_valid=_getbit(val, 2),
            tilt_valid=_getbit(val, 3),
            # bit 4 is unused
            vel=_getbit(val, 5),
            amp=_getbit(val, 6),
            corr=_getbit(val, 7),
            alt=_getbit(val, 8),
            alt_raw=_getbit(val, 9),
            ast=_getbit(val, 10),
            echo=_getbit(val, 11),
            ahrs=_getbit(val, 12),
            p_gd=_getbit(val, 13),
            std=_getbit(val, 14),
            # bit 15 is unused
        )
    elif mode == 'bt':
        return dict(
            press_valid=_getbit(val, 0),
            temp_valid=_getbit(val, 1),
            compass_valid=_getbit(val, 2),
            tilt_valid=_getbit(val, 3),
            # bit 4 is unused
            vel=_getbit(val, 5),
            # bits 6-7 unused
            dist=_getbit(val, 8),
            fom=_getbit(val, 9),
            ahrs=_getbit(val, 10),
            # bits 10-15 unused
        )


def _status02data(val):
    # This is detailed in the 6.1.2 of the Nortek Signature
    # Integrators Guide (2017)
    bi = _BitIndexer(val)
    out = {}
    if any(bi[15]):  # 'status0_in_use'
        out['proc_idle_less_3pct'] = bi[0]
        out['proc_idle_less_6pct'] = bi[1]
        out['proc_idle_less_12pct'] = bi[2]

    return out


def _status2data(val):
    # This is detailed in the 6.1.2 of the Nortek Signature
    # Integrators Guide (2017)
    bi = _BitIndexer(val)
    out = {}
    out['wakeup_state'] = bi[28:32]
    out['orient_up'] = bi[25:28]
    out['auto_orientation'] = bi[22:25]
    out['previous_wakeup_state'] = bi[18:22]
    out['low_volt_skip'] = bi[17]
    out['active_config'] = bi[16]
    out['echo_index'] = bi[12:16]
    out['telemetry_data'] = bi[11]
    out['boost_running'] = bi[10]
    out['echo_freq_bin'] = bi[5:10]
    # 2,3,4 unused
    out['bd_scaling'] = bi[1]  # if True: cm scaling of blanking dist
    # 0 unused
    return out


def _alt_status2data(val):
    # This is detailed in the 6.1.2 of the Nortek Signature
    # Integrators Guide (2017)
    bi = _BitIndexer(val)
    out = {}
    out['tilt_over_5deg'] = bi[0]
    out['tilt_over_10deg'] = bi[1]
    out['multibeam_alt'] = bi[2]
    out['n_beams_alt'] = bi[3:7]
    out['power_level_idx_alt'] = bi[7:10]

    return out


def _beams_cy_int2dict(val, id):
    """Convert the beams/coordinate-system bytes to a dict of values.
    """
    if id == 28:  # 0x1C (echosounder)
        return dict(n_cells=val)

    return dict(
        n_cells=val & (2 ** 10 - 1),
        cy=['ENU', 'XYZ', 'beam', None][val >> 10 & 3],
        n_beams=val >> 12)


def _isuniform(vec, exclude=[]):
    if len(exclude):
        return len(set(np.unique(vec)) - set(exclude)) <= 1
    return np.all(vec == vec[0])


def _collapse(vec, name=None, exclude=[]):
    """Check that the input vector is uniform, then collapse it to a
    single value, otherwise raise a warning.
    """
    if _isuniform(vec):
        return vec[0]
    elif _isuniform(vec, exclude=exclude):
        return list(set(np.unique(vec)) - set(exclude))[0]
    else:
        uniq, idx, counts = np.unique(
            vec, return_index=True, return_counts=True)

        if all(e == counts[0] for e in counts):
            val = max(vec)  # pings saved out of order, but equal # of pings
        else:
            val = vec[idx[np.argmax(counts)]]

        if not set(uniq) == set([0, val]) and set(counts) == set([1, np.max(counts)]):
            # warn when the 'wrong value' is not just a single zero.
            warnings.warn("The variable {} is expected to be uniform, but it is not.\n"
                          "Values found: {} (counts: {}).\n"
                          "Using the most common value: {}".format(
                              name, list(uniq), list(counts), val))
        return val


def _calc_config(index):
    """Calculate the configuration information (e.g., number of pings,
    number of beams, struct types, etc.) from the index data.

    Returns
    =======
    config : dict
        A dict containing the key information for initializing arrays.
    """
    ids = np.unique(index['ID'])
    config = {}
    for id in ids:
        if id not in [21, 23, 24, 26, 28]:
            continue
        if id == 23:
            type = 'bt'
        else:
            type = 'burst'
        inds = index['ID'] == id
        _config = index['config'][inds]
        _beams_cy = index['beams_cy'][inds]
        # Check that these variables are consistent
        if not _isuniform(_config):
            raise Exception("config are not identical for id: 0x{:X}."
                            .format(id))
        if not _isuniform(_beams_cy):
            raise Exception("beams_cy are not identical for id: 0x{:X}."
                            .format(id))
        # Now that we've confirmed they are the same:
        config[id] = _headconfig_int2dict(_config[0], mode=type)
        config[id].update(_beams_cy_int2dict(_beams_cy[0], id))
        config[id]['_config'] = _config[0]
        config[id]['_beams_cy'] = _beams_cy[0]
        config[id]['type'] = type
        config[id].pop('cy', None)
    return config



import numpy as np
import xarray as xr
from .binned import TimeBinner
from .time import dt642epoch, dt642date
from .rotate.api import rotate2, set_declination, set_inst2head_rotmat
from .io.api import save
from .tools.psd import coherence, phase_angle
from .tools.misc import slice1d_along_axis


@xr.register_dataset_accessor('velds')  # 'vel dataset'
class Velocity():
    """
    All ADCP and ADV xarray datasets wrap this base class.
    The turbulence-related attributes defined within this class 
    assume that the  ``'tke_vec'`` and ``'stress_vec'`` data entries are 
    included in the dataset. These are typically calculated using a
    :class:`VelBinner` tool, but the method for calculating these
    variables can depend on the details of the measurement
    (instrument, it's configuration, orientation, etc.).

    See Also
    ========
    :class:`VelBinner`
    """

    ########
    # Major components of the dolfyn-API

    def rotate2(self, out_frame='earth', inplace=True):
        """
        Rotate the dataset to a new coordinate system.

        Parameters
        ----------
        out_frame : string {'beam', 'inst', 'earth', 'principal'}
          The coordinate system to rotate the data into.
        inplace : bool (default: True)
          When True the existing data object is modified. When False
          a copy is returned.

        Returns
        -------
        ds : xarray.Dataset or None
          Returns the rotated dataset **when ``inplace=False``**, otherwise
          returns None.

        Notes
        -----
        - This function rotates all variables in ``ds.attrs['rotate_vars']``.

        - To rotate to the 'principal' frame, a value of
          ``ds.attrs['principal_heading']`` must exist. The function
          :func:`calc_principal_heading <dolfyn.calc_principal_heading>`
          is recommended for this purpose, e.g.::

              ds.attrs['principal_heading'] = dolfyn.calc_principal_heading(ds['vel'].mean(range))

          where here we are using the depth-averaged velocity to calculate
          the principal direction.
        """

        return rotate2(self.ds, out_frame, inplace)

    def set_declination(self, declin, inplace=True):
        """
        Set the magnetic declination

        Parameters
        ----------
        declination : float
          The value of the magnetic declination in degrees (positive
          values specify that Magnetic North is clockwise from True North)
        inplace : bool (default: True)
          When True the existing data object is modified. When False
          a copy is returned.

        Returns
        -------
        ds : xarray.Dataset or None
          Returns the rotated dataset **when ``inplace=False``**, otherwise
          returns None.

        Notes
        -----
        This method modifies the data object in the following ways:

        - If the dataset is in the *earth* reference frame at the time of
        setting declination, it will be rotated into the "*True-East*,
        *True-North*, Up" (hereafter, ETU) coordinate system

        - ``dat['orientmat']`` is modified to be an ETU to
        instrument (XYZ) rotation matrix (rather than the magnetic-ENU to
        XYZ rotation matrix). Therefore, all rotations to/from the 'earth'
        frame will now be to/from this ETU coordinate system.

        - The value of the specified declination will be stored in
        ``dat.attrs['declination']``

        - ``dat['heading']`` is adjusted for declination
        (i.e., it is relative to True North).

        - If ``dat.attrs['principal_heading']`` is set, it is
        adjusted to account for the orientation of the new 'True'
        earth coordinate system (i.e., calling set_declination on a
        data object in the principal coordinate system, then calling
        dat.rotate2('earth') will yield a data object in the new
        'True' earth coordinate system)
        """

        return set_declination(self.ds, declin, inplace)

    def set_inst2head_rotmat(self, rotmat, inplace=True):
        """
        Set the instrument to head rotation matrix for the Nortek ADV if it
        hasn't already been set through a '.userdata.json' file.

        Parameters
        ----------
        rotmat : float
            3x3 rotation matrix
        inplace : bool (default: True)
            When True the existing data object is rotated. When False
            a copy is returned that is rotated.

        Returns
        -------
        ds : xarray.Dataset or None
          Returns the rotated dataset **when ``inplace=False``**, otherwise
          returns None.

        Notes
        -----
        If the data object is in earth or principal coords, it is first
        rotated to 'inst' before assigning inst2head_rotmat, it is then
        rotated back to the coordinate system in which it was input. This
        way the inst2head_rotmat gets applied correctly (in inst
        coordinate system).
        """

        return set_inst2head_rotmat(self.ds, rotmat, inplace)

    def save(self, filename, **kwargs):
        """
        Save the data object (underlying xarray dataset) as netCDF (.nc).

        Parameters
        ----------
        filename : str
            Filename and/or path with the '.nc' extension
        **kwargs : these are passed directly to :func:`xarray.Dataset.to_netcdf`.

        Notes
        -----
        See |dlfn|'s :func:`save <dolfyn.io.api.save>` function for
        additional details.
        """

        save(self.ds, filename, **kwargs)

    ########
    # Magic methods of the API

    def __init__(self, ds, *args, **kwargs):
        self.ds = ds

    def __getitem__(self, key):
        return self.ds[key]

    def __contains__(self, val):
        return val in self.ds

    def __repr__(self, ):
        time_string = '{:.2f} {} (started: {})'
        if ('time' not in self or dt642epoch(self['time'][0]) < 1):
            time_string = '-->No Time Information!<--'
        else:
            tm = self['time'][[0, -1]].values
            dt = dt642date(tm[0])[0]
            delta = (dt642epoch(tm[-1]) -
                     dt642epoch(tm[0])) / (3600 * 24)  # days
            if delta > 1:
                units = 'days'
            elif delta * 24 > 1:
                units = 'hours'
                delta *= 24
            elif delta * 24 * 60 > 1:
                delta *= 24 * 60
                units = 'minutes'
            else:
                delta *= 24 * 3600
                units = 'seconds'
            try:
                time_string = time_string.format(delta, units,
                                                 dt.strftime('%b %d, %Y %H:%M'))
            except AttributeError:
                time_string = '-->Error in time info<--'

        p = self.ds.attrs
        t_shape = self['time'].shape
        if len(t_shape) > 1:
            shape_string = '({} bins, {} pings @ {}Hz)'.format(
                t_shape[0], t_shape, p.get('fs'))
        else:
            shape_string = '({} pings @ {}Hz)'.format(
                t_shape[0], p.get('fs', '??'))
        _header = ("<%s data object>: "
                   " %s %s\n"
                   "  . %s\n"
                   "  . %s-frame\n"
                   "  . %s\n" %
                   (p.get('inst_type'),
                    self.ds.attrs['inst_make'], self.ds.attrs['inst_model'],
                    time_string,
                    p.get('coord_sys'),
                    shape_string))
        _vars = '  Variables:\n'

        # Specify which variable show up in this view here.
        # * indicates a wildcard
        # This list also sets the display order.
        # Only the first 12 matches are displayed.
        show_vars = ['time*', 'vel*', 'range', 'range_echo',
                     'orientmat', 'heading', 'pitch', 'roll',
                     'temp', 'press*', 'amp*', 'corr*',
                     'accel', 'angrt', 'mag',
                     'echo',
                     ]
        n = 0
        for v in show_vars:
            if n > 12:
                break
            if v.endswith('*'):
                v = v[:-1]  # Drop the '*'
                for nm in self.variables:
                    if n > 12:
                        break
                    if nm.startswith(v):
                        n += 1
                        _vars += '  - {} {}\n'.format(nm, self.ds[nm].dims)
            elif v in self.ds:
                _vars += '  - {} {}\n'.format(v, self.ds[v].dims)
        if n < len(self.variables):
            _vars += '  ... and others (see `<obj>.variables`)\n'
        return _header + _vars

    ######
    # Duplicate valuable xarray properties here.
    @property
    def variables(self, ):
        """A sorted list of the variable names in the dataset."""
        return sorted(self.ds.variables)

    @property
    def attrs(self, ):
        """The attributes in the dataset."""
        return self.ds.attrs

    @property
    def coords(self, ):
        """The coordinates in the dataset."""
        return self.ds.coords

    ######
    # A bunch of DOLfYN specific properties
    @property
    def u(self,):
        """The first velocity component.

        This is simply a shortcut to self['vel'][0]. Therefore,
        depending on the coordinate system of the data object
        (self.attrs['coord_sys']), it is:

        - beam:      beam1
        - inst:      x
        - earth:     east
        - principal: streamwise
        """
        return self.ds['vel'][0]

    @property
    def v(self,):
        """The second velocity component.

        This is simply a shortcut to self['vel'][1]. Therefore,
        depending on the coordinate system of the data object
        (self.attrs['coord_sys']), it is:

        - beam:      beam2
        - inst:      y
        - earth:     north
        - principal: cross-stream
        """
        return self.ds['vel'][1]

    @property
    def w(self,):
        """The third velocity component.
        This is simply a shortcut to self['vel'][2]. Therefore,
        depending on the coordinate system of the data object
        (self.attrs['coord_sys']), it is:
        - beam:      beam3
        - inst:      z
        - earth:     up
        - principal: up
        """
        return self.ds['vel'][2]

    @property
    def U(self,):
        """Horizontal velocity as a complex quantity
        """
        return xr.DataArray(
            (self.u + self.v * 1j),
            attrs={'units': 'm/s',
                   'description': 'horizontal velocity (complex)'})

    @property
    def U_mag(self,):
        """Horizontal velocity magnitude
        """
        return xr.DataArray(
            np.abs(self.U),
            attrs={'units': 'm/s',
                   'description': 'horizontal velocity magnitude'})

    @property
    def U_dir(self,):
        """Angle of horizontal velocity vector, degrees counterclockwise from
        X/East/streamwise. Direction is 'to', as opposed to 'from'.
        """
        # Convert from radians to degrees
        angle = np.angle(self.U)*(180/np.pi)

        return xr.DataArray(angle,
                            dims=self.U.dims,
                            coords=self.U.coords,
                            attrs={'units': 'deg',
                                   'description': 'horizontal velocity flow direction, CCW from X/East/streamwise'})

    @property
    def E_coh(self,):
        """Coherent turbulent energy

        Niel Kelley's 'coherent turbulence energy', which is the RMS
        of the Reynold's stresses.

        See: NREL Technical Report TP-500-52353
        """
        E_coh = (self.upwp_**2 + self.upvp_**2 + self.vpwp_**2) ** (0.5)

        return xr.DataArray(E_coh,
                            coords={'time': self.ds['stress_vec'].time},
                            dims=['time'],
                            attrs={'units': self.ds['stress_vec'].units},
                            name='E_coh')

    @property
    def I_tke(self, thresh=0):
        """Turbulent kinetic energy intensity.

        Ratio of sqrt(tke) to horizontal velocity magnitude.
        """
        I_tke = np.ma.masked_where(self.U_mag < thresh,
                                   np.sqrt(2 * self.tke) / self.U_mag)
        return xr.DataArray(I_tke.data,
                            coords=self.U_mag.coords,
                            dims=self.U_mag.dims,
                            attrs={'units': '% [0,1]'},
                            name='TKE intensity')

    @property
    def I(self, thresh=0):
        """Turbulence intensity.

        Ratio of standard deviation of horizontal velocity
        to horizontal velocity magnitude.
        """
        I = np.ma.masked_where(self.U_mag < thresh,
                               self.ds['U_std'] / self.U_mag)
        return xr.DataArray(I.data,
                            coords=self.U_mag.coords,
                            dims=self.U_mag.dims,
                            attrs={'units': '% [0,1]'},
                            name='turbulence intensity')

    @property
    def tke(self,):
        """Turbulent kinetic energy (sum of the three components)
        """
        tke = self.ds['tke_vec'].sum('tke') / 2
        tke.name = 'TKE'
        tke.attrs['units'] = self.ds['tke_vec'].units
        return tke

    @property
    def upvp_(self,):
        """u'v'bar Reynolds stress
        """
        return self.ds['stress_vec'].sel(tau="upvp_")

    @property
    def upwp_(self,):
        """u'w'bar Reynolds stress
        """
        return self.ds['stress_vec'].sel(tau="upwp_")

    @property
    def vpwp_(self,):
        """v'w'bar Reynolds stress
        """
        return self.ds['stress_vec'].sel(tau="vpwp_")

    @property
    def upup_(self,):
        """u'u'bar component of the tke
        """
        return self.ds['tke_vec'].sel(tke="upup_")

    @property
    def vpvp_(self,):
        """v'v'bar component of the tke
        """
        return self.ds['tke_vec'].sel(tke="vpvp_")

    @property
    def wpwp_(self,):
        """w'w'bar component of the tke
        """
        return self.ds['tke_vec'].sel(tke="wpwp_")


class VelBinner(TimeBinner):
    """
    This is the base binning (averaging) tool.
    All |dlfn| binning tools derive from this base class.

    Examples
    ========
    The VelBinner class is used to compute averages and turbulence
    statistics from 'raw' (not averaged) ADV or ADP measurements, for
    example::

        # First read or load some data.
        rawdat = dolfyn.read_example('BenchFile01.ad2cp')

        # Now initialize the averaging tool:
        binner = dolfyn.VelBinner(n_bin=600, fs=rawdat.fs)

        # This computes the basic averages
        avg = binner.do_avg(rawdat)
    """

    # This defines how cross-spectra and stresses are computed.
    _cross_pairs = [(0, 1), (0, 2), (1, 2)]

    def do_avg(self, raw_ds, out_ds=None, names=None, noise=[0, 0, 0]):
        """
        Bin the dataset and calculate the ensemble averages of each 
        variable.

        Parameters
        ----------
        raw_ds : xarray.Dataset
           The raw data structure to be binned
        out_ds : xarray.Dataset
           The bin'd (output) data object to which averaged data is added.
        names : list of strings
           The names of variables to be averaged.  If `names` is None,
           all data in `raw_ds` will be binned.
        noise : list or numpy.ndarray
          Instrument's doppler noise in same units as velocity

        Returns
        -------
        out_ds : xarray.Dataset
          The new (or updated when out_ds is not None) dataset
          with the averages of all the variables in raw_ds.

        Raises
        ------
        AttributeError : when out_ds is supplied as input (not None)
        and the values in out_ds.attrs are inconsistent with
        raw_ds.attrs or the properties of this VelBinner (n_bin,
        n_fft, fs, etc.)

        Notes
        -----
        raw_ds.attrs are copied to out_ds.attrs. Inconsistencies
        between the two (when out_ds is specified as input) raise an
        AttributeError.
        """

        out_ds = self._check_ds(raw_ds, out_ds)

        if names is None:
            names = raw_ds.data_vars

        for ky in names:
            # set up dimensions and coordinates for Dataset
            dims_list = raw_ds[ky].dims
            coords_dict = {}
            for nm in dims_list:
                if 'time' in nm:
                    coords_dict[nm] = self.mean(raw_ds[ky][nm].values)
                else:
                    coords_dict[nm] = raw_ds[ky][nm].values

            # create Dataset
            if 'ensemble' not in ky:
                try:  # variables with time coordinate
                    out_ds[ky] = xr.DataArray(self.mean(raw_ds[ky].values),
                                              coords=coords_dict,
                                              dims=dims_list,
                                              attrs=raw_ds[ky].attrs
                                              ).astype('float32')
                except:  # variables not needing averaging
                    pass
            # Add standard deviation
            std = (np.nanstd(self.reshape(raw_ds.velds.U_mag.values),
                             axis=-1,
                             dtype=np.float64) - (noise[0] + noise[1])/2)
            out_ds['U_std'] = xr.DataArray(
                std.astype('float32'),
                dims=raw_ds.vel.dims[1:],
                attrs={'units': 'm/s',
                       'description': 'horizontal velocity std dev'})

        return out_ds

    def do_var(self, raw_ds, out_ds=None, names=None, suffix='_var'):
        """
        Bin the dataset and calculate the ensemble variances of each 
        variable. Complementary to `do_avg()`.

        Parameters
        ----------
        raw_ds : xarray.Dataset
           The raw data structure to be binned.
        out_ds : xarray.Dataset
           The binned (output) dataset to which variance data is added,
           nominally dataset output from `do_avg()`
        names : list of strings
           The names of variables of which to calculate variance.  If
           `names` is None, all data in `raw_ds` will be binned.

        Returns
        -------
        out_ds : xarray.Dataset
          The new (or updated when out_ds is not None) dataset
          with the variance of all the variables in raw_ds.

        Raises
        ------
        AttributeError : when out_ds is supplied as input (not None)
        and the values in out_ds.attrs are inconsistent with
        raw_ds.attrs or the properties of this VelBinner (n_bin,
        n_fft, fs, etc.)

        Notes
        -----
        raw_ds.attrs are copied to out_ds.attrs. Inconsistencies
        between the two (when out_ds is specified as input) raise an
        AttributeError.
        """

        out_ds = self._check_ds(raw_ds, out_ds)

        if names is None:
            names = raw_ds.data_vars

        for ky in names:
            # set up dimensions and coordinates for dataarray
            dims_list = raw_ds[ky].dims
            coords_dict = {}
            for nm in dims_list:
                if 'time' in nm:
                    coords_dict[nm] = self.mean(raw_ds[ky][nm].values)
                else:
                    coords_dict[nm] = raw_ds[ky][nm].values

            # create Dataset
            if 'ensemble' not in ky:
                try:  # variables with time coordinate
                    out_ds[ky+suffix] = xr.DataArray(self.var(raw_ds[ky].values),
                                                     coords=coords_dict,
                                                     dims=dims_list,
                                                     attrs=raw_ds[ky].attrs
                                                     ).astype('float32')
                except:  # variables not needing averaging
                    pass

        return out_ds

    def calc_coh(self, veldat1, veldat2, window='hann', debias=True,
                 noise=(0, 0), n_fft_coh=None, n_bin=None):
        """
        Calculate coherence between `veldat1` and `veldat2`.

        Parameters
        ----------
        veldat1 : xarray.DataArray
          The first (the longer, if applicable) raw dataArray of which to 
          calculate coherence
        veldat2 : xarray.DataArray
          The second (the shorter, if applicable) raw dataArray of which to 
          calculate coherence
        window : str
          String indicating the window function to use (default: 'hanning')
        noise : float
          The white-noise level of the measurement (in the same units
          as `veldat`).
        n_fft_coh : int
          n_fft of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner
        n_bin : int
          n_bin of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner

        Returns
        -------
        da : xarray.DataArray
          The coherence between signal veldat1 and veldat2.

        Notes
        -----
        The two velocity inputs do not have to be perfectly synchronized, but 
        they should have the same start and end timestamps.
        """

        if veldat1.size < veldat2.size:
            raise Exception(
                "veldat1 is shorter than veldat2. Please switch these inputs.")

        dat1 = veldat1.values
        dat2 = veldat2.values

        if n_fft_coh is None:
            n_fft = self.n_fft_coh
        else:
            n_fft = int(n_fft_coh)

        # want each slice to carry the same timespan
        n_bin2 = self._parse_nbin(n_bin)  # bins for shorter array
        n_bin1 = int(dat1.shape[-1]/(dat2.shape[-1]/n_bin2))

        oshp = self._outshape_fft(dat1.shape, n_fft=n_fft, n_bin=n_bin1)
        oshp[-2] = np.min([oshp[-2], int(dat2.shape[-1] // n_bin2)])
        out = np.empty(oshp, dtype=dat1.dtype)

        # The data is detrended in psd, so we don't need to do it here.
        dat1 = self.reshape(dat1, n_pad=n_fft, n_bin=n_bin1)
        dat2 = self.reshape(dat2, n_pad=n_fft, n_bin=n_bin2)

        for slc in slice1d_along_axis(out.shape, -1):
            out[slc] = coherence(dat1[slc], dat2[slc], n_fft,
                                 window=window, debias=debias,
                                 noise=noise)

        freq = self.calc_freq(self.fs, units='Hz', coh=True)

        # Get time from shorter vector
        dims_list, coords_dict = self._new_coords(veldat2)
        # tack on new coordinate
        dims_list.append('f')
        coords_dict['f'] = freq

        da = xr.DataArray(out.astype('float32'),
                          coords=coords_dict,
                          dims=dims_list)
        da['f'].attrs['units'] = 'Hz'

        return da

    def calc_phase_angle(self, veldat1, veldat2, window='hann',
                         n_fft_coh=None, n_bin=None):
        """
        Calculate the phase difference between two signals as a
        function of frequency (complimentary to coherence).

        Parameters
        ----------
        veldat1 : xarray.DataArray
          The first (the longer, if applicable) raw dataArray of which to 
          calculate phase angle
        veldat2 : xarray.DataArray
          The second (the shorter, if applicable) raw dataArray of which 
          to calculate phase angle
        window : str
          String indicating the window function to use (default: 'hanning').
        n_fft : int
          Number of elements per bin if 'None' is taken from VelBinner
        n_bin : int
          Number of elements per bin from veldat2 if 'None' is taken 
          from VelBinner

        Returns
        -------
        da : xarray.DataArray
          The phase difference between signal veldat1 and veldat2.

        Notes
        -----
        The two velocity inputs do not have to be perfectly synchronized, but 
        they should have the same start and end timestamps.
        """

        if veldat1.size < veldat2.size:
            raise Exception(
                "veldat1 is shorter than veldat2. Please switch these inputs.")

        dat1 = veldat1.values
        dat2 = veldat2.values

        if n_fft_coh is None:
            n_fft = self.n_fft_coh
        else:
            n_fft = int(n_fft_coh)

        # want each slice to carry the same timespan
        n_bin2 = self._parse_nbin(n_bin)  # bins for shorter array
        n_bin1 = int(dat1.shape[-1]/(dat2.shape[-1]/n_bin2))

        oshp = self._outshape_fft(dat1.shape, n_fft=n_fft, n_bin=n_bin1)
        oshp[-2] = np.min([oshp[-2], int(dat2.shape[-1] // n_bin2)])

        # The data is detrended in psd, so we don't need to do it here:
        dat1 = self.reshape(dat1, n_pad=n_fft, n_bin=n_bin1)
        dat2 = self.reshape(dat2, n_pad=n_fft, n_bin=n_bin2)
        out = np.empty(oshp, dtype='c{}'.format(dat2.dtype.itemsize * 2))

        for slc in slice1d_along_axis(out.shape, -1):
            # PSD's are computed in radian units:
            out[slc] = phase_angle(dat1[slc], dat2[slc], n_fft,
                                   window=window)

        freq = self.calc_freq(self.fs, units='Hz', coh=True)

        # Get time from shorter vector
        dims_list, coords_dict = self._new_coords(veldat2)
        # tack on new coordinate
        dims_list.append('f')
        coords_dict['f'] = freq

        da = xr.DataArray(out,
                          coords=coords_dict,
                          dims=dims_list)
        da['f'].attrs['units'] = 'Hz'

        return da

    def calc_acov(self, veldat, n_bin=None):
        """
        Calculate the auto-covariance of the raw-signal `veldat`

        Parameters
        ----------
        veldat : xarray.DataArray
          The raw dataArray of which to calculate auto-covariance
        n_bin : float
          Number of data elements to use

        Returns
        -------
        da : xarray.DataArray
          The auto-covariance of veldat

        Notes
        -----
        As opposed to calc_xcov, which returns the full
        cross-covariance between two arrays, this function only
        returns a quarter of the full auto-covariance. It computes the
        auto-covariance over half of the range, then averages the two
        sides (to return a 'quartered' covariance).

        This has the advantage that the 0 index is actually zero-lag.
        """

        indat = veldat.values

        n_bin = self._parse_nbin(n_bin)
        out = np.empty(self._outshape(indat.shape, n_bin=n_bin)[:-1] +
                       [int(n_bin // 4)], dtype=indat.dtype)
        dt1 = self.reshape(indat, n_pad=n_bin / 2 - 2)
        # Here we de-mean only on the 'valid' range:
        dt1 = dt1 - dt1[..., :, int(n_bin // 4):
                        int(-n_bin // 4)].mean(-1)[..., None]
        dt2 = self.demean(indat)
        se = slice(int(n_bin // 4) - 1, None, 1)
        sb = slice(int(n_bin // 4) - 1, None, -1)
        for slc in slice1d_along_axis(dt1.shape, -1):
            tmp = np.correlate(dt1[slc], dt2[slc], 'valid')
            # The zero-padding in reshape means we compute coherence
            # from one-sided time-series for first and last points.
            if slc[-2] == 0:
                out[slc] = tmp[se]
            elif slc[-2] == dt2.shape[-2] - 1:
                out[slc] = tmp[sb]
            else:
                # For the others we take the average of the two sides.
                out[slc] = (tmp[se] + tmp[sb]) / 2

        dims_list, coords_dict = self._new_coords(veldat)
        # tack on new coordinate
        dims_list.append('lag')
        coords_dict['lag'] = np.arange(n_bin//4)

        da = xr.DataArray(out.astype('float32'),
                          coords=coords_dict,
                          dims=dims_list,)
        da['lag'].attrs['units'] = 'timestep'

        return da

    def calc_xcov(self, veldat1, veldat2, npt=1,
                  n_bin=None, normed=False):
        """
        Calculate the cross-covariance between arrays veldat1 and veldat2

        Parameters
        ----------
        veldat1 : xarray.DataArray
          The first raw dataArray of which to calculate cross-covariance
        veldat2 : xarray.DataArray
          The second raw dataArray of which to calculate cross-covariance
        npt : int
          Number of timesteps (lag) to calculate covariance
        n_fft : int
          n_fft of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner
        n_bin : int
          n_bin of veldat2, number of elements per bin if 'None' is taken 
          from VelBinner

        Returns
        -------
        da : xarray.DataArray
          The cross-covariance between signal veldat1 and veldat2.

        Notes
        -----
        The two velocity inputs must be the same length
        """

        dat1 = veldat1.values
        dat2 = veldat2.values

        # want each slice to carry the same timespan
        n_bin2 = self._parse_nbin(n_bin)
        n_bin1 = int(dat1.shape[-1]/(dat2.shape[-1]/n_bin2))

        shp = self._outshape(dat1.shape, n_bin=n_bin1)
        shp[-2] = min(shp[-2], self._outshape(dat2.shape, n_bin=n_bin2)[-2])

        # reshape dat1 to be the same size as dat2
        out = np.empty(shp[:-1] + [npt], dtype=dat1.dtype)
        tmp = int(n_bin2) - int(n_bin1) + npt
        dt1 = self.reshape(dat1, n_pad=tmp-1, n_bin=n_bin1)

        # Note here I am demeaning only on the 'valid' range:
        dt1 = dt1 - dt1[..., :, int(tmp // 2)
                                    :int(-tmp // 2)].mean(-1)[..., None]
        # Don't need to pad the second variable:
        dt2 = self.demean(dat2, n_bin=n_bin2)

        for slc in slice1d_along_axis(shp, -1):
            out[slc] = np.correlate(dt1[slc], dt2[slc], 'valid')
        if normed:
            out /= (self.std(dat1, n_bin=n_bin1)[..., :shp[-2]] *
                    self.std(dat2, n_bin=n_bin2)[..., :shp[-2]] *
                    n_bin2)[..., None]

        dims_list, coords_dict = self._new_coords(veldat1)
        # tack on new coordinate
        dims_list.append('dt')
        coords_dict['dt'] = np.arange(npt)

        da = xr.DataArray(out.astype('float32'),
                          coords=coords_dict,
                          dims=dims_list)
        da['dt'].attrs['units'] = 'timestep'

        return da

    def calc_tke(self, veldat, noise=[0, 0, 0], detrend=True):
        """
        Calculate the turbulent kinetic energy (TKE) (variances 
        of u,v,w).

        Parameters
        ----------
        veldat : xarray.DataArray
            Velocity data array from ADV or single beam from ADCP. 
            The last dimension is assumed to be time.
        noise : float
            a three-element vector of the noise levels of the
            velocity data for ach component of velocity.
        detrend : bool (default: False)
            detrend the velocity data (True), or simply de-mean it
            (False), prior to computing tke. Note: the psd routines
            use detrend, so if you want to have the same amount of
            variance here as there use ``detrend=True``.

        Returns
        -------
        ds : xarray.DataArray
            dataArray containing u'u'_, v'v'_ and w'w'_
        """

        if 'dir' in veldat.dims:
            # will error for ADCP 4-beam, but not for single beam
            vel = veldat.values
        else:  # for single beam input
            vel = veldat.values

        if detrend:
            vel = self.detrend(vel)
        else:
            vel = self.demean(vel)

        if 'time_b5' in veldat.dims:
            time = self.mean(veldat.time_b5.values)
        else:
            time = self.mean(veldat.time.values)

        out = np.nanmean(vel**2, -1,
                         dtype=np.float64,
                         ).astype('float32')

        out[0] -= noise[0] ** 2
        out[1] -= noise[1] ** 2
        out[2] -= noise[2] ** 2

        da = xr.DataArray(out.astype('float32'),
                          dims=veldat.dims,
                          attrs={'units': 'm^2/^2'})

        if 'dir' in veldat.dims:
            da = da.rename({'dir': 'tke'})
            da = da.assign_coords({'tke': ["upup_", "vpvp_", "wpwp_"],
                                   'time': time})
        else:
            if 'time_b5' in veldat.dims:
                da = da.assign_coords({'time_b5': time})
            else:
                da = da.assign_coords({'time': time})

        return da

    def calc_psd(self, veldat,
                 freq_units='rad/s',
                 fs=None,
                 window='hann',
                 noise=[0, 0, 0],
                 n_bin=None, n_fft=None, n_pad=None,
                 step=None):
        """
        Calculate the power spectral density of velocity.

        Parameters
        ----------
        veldat : xr.DataArray
          The raw velocity data (of dims 'dir' and 'time').
        freq_units : string
          Frequency units of the returned spectra in either Hz or rad/s 
          (`f` or :math:`\\omega`)
        fs : float (optional)
          The sample rate (default: from the binner).
        window : string or array
          Specify the window function.
        noise : list(3 floats) (optional)
          Noise level of each component's velocity measurement
          (default 0).
        n_bin : int (optional)
          The bin-size (default: from the binner).
        n_fft : int (optional)
          The fft size (default: from the binner).
        n_pad : int (optional)
          The number of values to pad with zero (default: 0)
        step : int (optional)
          Controls amount of overlap in fft (default: the step size is
          chosen to maximize data use, minimize nens, and have a
          minimum of 50% overlap.).

        Returns
        -------
        psd : xarray.DataArray (3, M, N_FFT)
          The spectra in the 'u', 'v', and 'w' directions.
        """

        if 'time_b5' in veldat.dims:
            time = self.mean(veldat.time_b5.values)
            time_str = 'time_b5'
        else:
            time = self.mean(veldat.time.values)
            time_str = 'time'
        fs = self._parse_fs(fs)
        n_fft = self._parse_nfft(n_fft)
        veldat = veldat.values

        # Create frequency vector, also checks whether using f or omega
        freq = self.calc_freq(units=freq_units)
        if 'rad' in freq_units:
            fs = 2*np.pi*fs
            freq_units = 'rad/s'
            units = 'm^2/s/rad'
        else:
            freq_units = 'Hz'
            units = 'm^2/s^2/Hz'

        # Spectra, if input is full velocity or a single array
        if len(veldat.shape) == 2:
            assert veldat.shape[0] == 3, "Function can only handle 1D or 3D arrays." \
                " If ADCP data, please select a specific depth bin."

            out = np.empty(self._outshape_fft(veldat[:3].shape),
                           dtype=np.float32)
            for idx in range(3):
                out[idx] = self.calc_psd_base(veldat[idx], fs=fs, noise=noise[idx],
                                              window=window, n_bin=n_bin,
                                              n_pad=n_pad, n_fft=n_fft, step=step)
            coords = {'S': ['Sxx', 'Syy', 'Szz'], time_str: time, 'freq': freq}
            dims = ['S', time_str, 'freq']
        else:
            out = self.calc_psd_base(veldat, fs=fs, noise=noise[0], window=window,
                                     n_bin=n_bin, n_pad=n_pad, n_fft=n_fft, step=step)
            coords = {time_str: time, 'freq': freq}
            dims = [time_str, 'freq']

        psd = xr.DataArray(out.astype('float32'),
                           coords=coords,
                           dims=dims,
                           attrs={'units': units, 'n_fft': n_fft})
        psd['freq'].attrs['units'] = freq_units

        return psd



import numpy as np
from numpy.linalg import det, inv
from scipy.spatial.transform import Rotation as R


def _make_model(ds):
    """The make and model of the instrument that collected the data
    in this data object.
    """
    return '{} {}'.format(ds.attrs['inst_make'],
                          ds.attrs['inst_model']).lower()


def _check_rotmat_det(rotmat, thresh=1e-3):
    """Check that the absolute error of the determinant is small.

          abs(det(rotmat) - 1) < thresh

    Returns a boolean array.
    """
    if rotmat.ndim > 2:
        rotmat = np.transpose(rotmat)
    return np.abs(det(rotmat) - 1) < thresh


def _set_coords(ds, ref_frame, forced=False):
    """
    Checks the current reference frame and adjusts xarray coords/dims 
    as necessary.
    Makes sure assigned dataarray coordinates match what DOLfYN is reading in.

    """
    make = _make_model(ds)

    XYZ = ['X', 'Y', 'Z']
    ENU = ['E', 'N', 'U']
    beam = list(range(1, ds['vel'].shape[0]+1))
    principal = ['streamwise', 'x-stream', 'vert']

    # check make/model
    if 'rdi' in make:
        inst = ['X', 'Y', 'Z', 'err']
        earth = ['E', 'N', 'U', 'err']
        princ = ['streamwise', 'x-stream', 'vert', 'err']

    elif 'nortek' in make:
        if 'signature' in make or 'ad2cp' in make:
            inst = ['X', 'Y', 'Z1', 'Z2']
            earth = ['E', 'N', 'U1', 'U2']
            princ = ['streamwise', 'x-stream', 'vert1', 'vert2']

        else:  # AWAC or Vector
            inst = XYZ
            earth = ENU
            princ = principal

    orient = {'beam': beam, 'inst': inst, 'ship': inst, 'earth': earth,
              'principal': princ}
    orientIMU = {'beam': XYZ, 'inst': XYZ, 'ship': XYZ, 'earth': ENU,
                 'principal': principal}

    if forced:
        ref_frame += '-forced'

    # update 'orient' and 'orientIMU' dimensions
    ds['dir'] = orient[ref_frame]
    if hasattr(ds, 'dirIMU'):
        ds['dirIMU'] = orientIMU[ref_frame]
    ds['dir'].attrs['ref_frame'] = ref_frame
    ds.attrs['coord_sys'] = ref_frame

    # These are essentially one extra line to scroll through
    tag = ['', '_echo', '_bt']
    for tg in tag:
        if hasattr(ds, 'coord_sys_axes'+tg):
            ds.attrs.pop('coord_sys_axes'+tg)

    return ds


def _beam2inst(dat, reverse=False, force=False):
    """Rotate velocities from beam to instrument coordinates.

    Parameters
    ----------
    dat : xarray.Dataset
        The ADCP dataset
    reverse : bool (default: False)
        If True, this function performs the inverse rotation (inst->beam).
    force : bool (default: False), or list
        When true do not check which coordinate system the data is in
        prior to performing this rotation. When forced-rotations are
        applied, the string '-forced!' is appended to the
        dat.props['coord_sys'] string. If force is a list, it contains
        a list of variables that should be rotated (rather than the
        default values in adpo.props['rotate_vars']).

    """
    if not force:
        if not reverse and dat.coord_sys.lower() != 'beam':
            raise ValueError('The input must be in beam coordinates.')
        if reverse and dat.coord_sys != 'inst':
            raise ValueError('The input must be in inst coordinates.')

    rotmat = dat['beam2inst_orientmat']

    if isinstance(force, (list, set, tuple)):
        # You can force a distinct set of variables to be rotated by
        # specifying it here.
        rotate_vars = force
    else:
        rotate_vars = [ky for ky in dat.rotate_vars if dat[ky].shape[0] == rotmat.shape[0]]

    cs = 'inst'
    if reverse:
        # Can't use transpose because rotation is not between
        # orthogonal coordinate systems
        rotmat = inv(rotmat)
        cs = 'beam'
    for ky in rotate_vars:
        dat[ky].values = np.einsum('ij,j...->i...', rotmat, dat[ky].values)

    if force:
        dat = _set_coords(dat, cs, forced=True)
    else:
        dat = _set_coords(dat, cs)

    return dat


def euler2orient(heading, pitch, roll, units='degrees'):
    """
    Calculate the orientation matrix from DOLfYN-defined euler angles.

    This function is not likely to be called during data processing since it requires
    DOLfYN-defined euler angles. It is intended for testing DOLfYN.

    The matrices H, P, R are the transpose of the matrices for rotation about z, y, x
    as shown here https://en.wikipedia.org/wiki/Rotation_matrix. The transpose is used
    because in DOLfYN the orientation matrix is organized for 
    rotation from EARTH --> INST, while the wiki's matrices are organized for 
    rotation from INST --> EARTH.

    Parameters
    ----------
    heading : |np.ndarray| (Nt)
      The heading angle.
    pitch : |np.ndarray| (Nt)
      The pitch angle.
    roll : |np.ndarray| (Nt)
      The roll angle.
    units : str {'degrees' (default), 'radians'}

    Returns
    =======
    omat : |np.ndarray| (3x3xNt)
      The orientation matrix of the data. The returned orientation
      matrix obeys the following conventions:

       - a "ZYX" rotation order. That is, these variables are computed
         assuming that rotation from the earth -> instrument frame happens
         by rotating around the z-axis first (heading), then rotating
         around the y-axis (pitch), then rotating around the x-axis (roll). 
         Note this requires matrix multiplication in the reverse order.

       - heading is defined as the direction the x-axis points, positive
         clockwise from North (this is *opposite* the right-hand-rule
         around the Z-axis), range 0-360 degrees.

       - pitch is positive when the x-axis pitches up (this is *opposite* the
         right-hand-rule around the Y-axis)

       - roll is positive according to the right-hand-rule around the
         instrument's x-axis

    """
    if units.lower() == 'degrees':
        pitch = np.deg2rad(pitch)
        roll = np.deg2rad(roll)
        heading = np.deg2rad(heading)
    elif units.lower() == 'radians':
        pass
    else:
        raise Exception("Invalid units")

    # Converts the DOLfYN-defined heading to one that follows the right-hand-rule
    # reports heading as rotation of the y-axis positive counterclockwise from North
    heading = np.pi / 2 - heading

    # Converts the DOLfYN-defined pitch to one that follows the right-hand-rule.
    pitch = -pitch

    ch = np.cos(heading)
    sh = np.sin(heading)
    cp = np.cos(pitch)
    sp = np.sin(pitch)
    cr = np.cos(roll)
    sr = np.sin(roll)
    zero = np.zeros_like(sr)
    one = np.ones_like(sr)

    H = np.array(
        [[ch, sh, zero],
         [-sh, ch, zero],
         [zero, zero, one], ])
    P = np.array(
        [[cp, zero, -sp],
         [zero, one, zero],
         [sp, zero, cp], ])
    R = np.array(
        [[one, zero, zero],
         [zero, cr, sr],
         [zero, -sr, cr], ])

    return np.einsum('ij...,jk...,kl...->il...', R, P, H)


def orient2euler(omat):
    """
    Calculate DOLfYN-defined euler angles from the orientation matrix.

    Parameters
    ----------
    omat : |np.ndarray|
      The orientation matrix

    Returns
    -------
    heading : |np.ndarray|
      The heading angle. Heading is defined as the direction the x-axis points,
      positive clockwise from North (this is *opposite* the right-hand-rule
      around the Z-axis), range 0-360 degrees.
    pitch : np.ndarray
      The pitch angle (degrees). Pitch is positive when the x-axis 
      pitches up (this is *opposite* the right-hand-rule around the Y-axis).
    roll : np.ndarray
      The roll angle (degrees). Roll is positive according to the 
      right-hand-rule around the instrument's x-axis.

    """

    if isinstance(omat, np.ndarray) and \
            omat.shape[:2] == (3, 3):
        pass
    elif hasattr(omat, 'orientmat'):
        omat = omat['orientmat'].values

    # Note: orientation matrix is earth->inst unless supplied by an external IMU
    hh = np.rad2deg(np.arctan2(omat[0, 0], omat[0, 1]))
    hh %= 360
    return (
        # heading
        hh,
        # pitch
        np.rad2deg(np.arcsin(omat[0, 2])),
        # roll
        np.rad2deg(np.arctan2(omat[1, 2], omat[2, 2])),
    )


def quaternion2orient(quaternions):
    """
    Calculate orientation from Nortek AHRS quaternions, where q = [W, X, Y, Z] 
    instead of the standard q = [X, Y, Z, W] = [q1, q2, q3, q4]

    Parameters
    ----------
    quaternions : xarray.DataArray
        Quaternion dataArray from the raw dataset

    Returns
    -------
    orientmat : |np.ndarray|
        The inst2earth rotation maxtrix as calculated from the quaternions

    See Also
    --------
    scipy.spatial.transform.Rotation

    """
    omat = type(quaternions)(np.empty((3, 3, quaternions.time.size)))
    omat = omat.rename({'dim_0': 'earth', 'dim_1': 'inst', 'dim_2': 'time'})

    for i in range(quaternions.time.size):
        r = R.from_quat([quaternions.isel(q=1, time=i),
                         quaternions.isel(q=2, time=i),
                         quaternions.isel(q=3, time=i),
                         quaternions.isel(q=0, time=i)])
        omat[..., i] = r.as_matrix()

    # quaternions in inst2earth reference frame, need to rotate to earth2inst
    omat.values = np.rollaxis(omat.values, 1)

    xyz = ['X', 'Y', 'Z']
    enu = ['E', 'N', 'U']
    return omat.assign_coords({'inst': xyz, 'earth': enu, 'time': quaternions.time})



import numpy as np
import xarray as xr
import json
import os
import warnings


def _abspath(fname):
    return os.path.abspath(os.path.expanduser(fname))


def _get_filetype(fname):
    """Detects whether the file is a Nortek, Signature (Nortek), or RDI
    file by reading the first few bytes of the file.

    Returns
    =======
       None - Doesn't match any known pattern
       'signature' - for Nortek signature files
       'nortek' - for Nortek (Vec, AWAC) files
       'RDI' - for RDI files
       '<GIT-LFS pointer> - if the file looks like a GIT-LFS pointer.
    """

    with open(fname, 'rb') as rdr:
        bytes = rdr.read(40)
    code = bytes[:2].hex()
    #print("{} - {}".format(fname.rsplit('/')[-1], bytes))
    if code in ['7f79', '7f7f']:
        return 'RDI'
    elif code in ['a50a']:
        return 'signature'
    elif code in ['a505']:
        # AWAC
        return 'nortek'
    elif bytes == b'version https://git-lfs.github.com/spec/':
        return '<GIT-LFS pointer>'
    else:
        return None


def _find_userdata(filename, userdata=True):
    # This function finds the file to read
    if userdata:
        for basefile in [filename.rsplit('.', 1)[0],
                         filename]:
            jsonfile = basefile + '.userdata.json'
            if os.path.isfile(jsonfile):
                return _read_userdata(jsonfile)

    elif isinstance(userdata, (str, )) or hasattr(userdata, 'read'):
        return _read_userdata(userdata)
    return {}


def _read_userdata(fname):
    """Reads a userdata.json file and returns the data it contains as a
    dictionary.
    """
    with open(fname) as data_file:
        data = json.load(data_file)
    for nm in ['body2head_rotmat', 'body2head_vec']:
        if nm in data:
            new_name = 'inst' + nm[4:]
            warnings.warn(
                f'{nm} has been deprecated, please change this to {new_name} \
                    in {fname}.')
            data[new_name] = data.pop(nm)
    if 'inst2head_rotmat' in data:
        if data['inst2head_rotmat'] in ['identity', 'eye', 1, 1.]:
            data['inst2head_rotmat'] = np.eye(3)
        else:
            data['inst2head_rotmat'] = np.array(data['inst2head_rotmat'])
    if 'inst2head_vec' in data and type(data['inst2head_vec']) != list:
        data['inst2head_vec'] = list(data['inst2head_vec'])

    return data


def _handle_nan(data):
    """Finds trailing nan's that cause issues in running the rotation 
    algorithms and deletes them.
    """
    nan = np.zeros(data['coords']['time'].shape, dtype=bool)
    l = data['coords']['time'].size

    if any(np.isnan(data['coords']['time'])):
        nan += np.isnan(data['coords']['time'])

    # Required for motion-correction algorithm
    var = ['accel', 'angrt', 'mag']
    for key in data['data_vars']:
        if any(val in key for val in var):
            shp = data['data_vars'][key].shape
            if shp[-1] == l:
                if len(shp) == 1:
                    if any(np.isnan(data['data_vars'][key])):
                        nan += np.isnan(data['data_vars'][key])
                elif len(shp) == 2:
                    if any(np.isnan(data['data_vars'][key][-1])):
                        nan += np.isnan(data['data_vars'][key][-1])
    trailing = np.cumsum(nan)[-1]

    if trailing > 0:
        data['coords']['time'] = data['coords']['time'][:-trailing]
        for key in data['data_vars']:
            if data['data_vars'][key].shape[-1] == l:
                data['data_vars'][key] = data['data_vars'][key][..., :-trailing]

    return data


def _create_dataset(data):
    """Creates an xarray dataset from dictionary created from binary
    readers.
    Direction 'dir' coordinates are set in `set_coords`
    """
    ds = xr.Dataset()
    inst = ['X', 'Y', 'Z']
    earth = ['E', 'N', 'U']
    beam = list(range(1, data['data_vars']['vel'].shape[0]+1))
    tag = ['_b5', '_echo', '_bt', '_gps', '_ast', '_sl']

    for key in data['data_vars']:
        # orientation matrices
        if 'mat' in key:
            if 'inst' in key:  # beam2inst & inst2head orientation matrices
                ds[key] = xr.DataArray(data['data_vars'][key],
                                       coords={'x': beam,
                                               'x*': beam},
                                       dims=['x', 'x*'])
            else:  # earth2inst orientation matrix
                if any(val in key for val in tag):
                    tg = '_' + key.rsplit('_')[-1]
                else:
                    tg = ''
                time = data['coords']['time'+tg]
                coords = {'earth': earth, 'inst': inst, 'time'+tg: time}
                dims = ['earth', 'inst', 'time'+tg]
                ds[key] = xr.DataArray(data['data_vars'][key], coords, dims)

        # quaternion units never change
        elif 'quaternions' in key:
            if any(val in key for val in tag):
                tg = '_' + key.rsplit('_')[-1]
            else:
                tg = ''
            ds[key] = xr.DataArray(data['data_vars'][key],
                                   coords={'q': ['w', 'x', 'y', 'z'],
                                           'time'+tg: data['coords']['time'+tg]},
                                   dims=['q', 'time'+tg])
        else:
            ds[key] = xr.DataArray(data['data_vars'][key])
            if key in data['units']:   # not all variables have units
                ds[key].attrs['units'] = data['units'][key]
            try:  # make sure ones with tags get units
                tg = '_' + key.rsplit('_')[-1]
                if any(val in key for val in tag):
                    ds[key].attrs['units'] = data['units'][key[:-len(tg)]]
            except:
                pass

            shp = data['data_vars'][key].shape
            vshp = data['data_vars']['vel'].shape
            l = len(shp)
            if l == 1:  # 1D variables
                if any(val in key for val in tag):
                    tg = '_' + key.rsplit('_')[-1]
                else:
                    tg = ''
                ds[key] = ds[key].rename({'dim_0': 'time'+tg})
                ds[key] = ds[key].assign_coords(
                    {'time'+tg: data['coords']['time'+tg]})

            elif l == 2:  # 2D variables
                if key == 'echo':
                    ds[key] = ds[key].rename({'dim_0': 'range_echo',
                                              'dim_1': 'time_echo'})
                    ds[key] = ds[key].assign_coords({'range_echo': data['coords']['range_echo'],
                                                     'time_echo': data['coords']['time_echo']})
                # ADV/ADCP instrument vector data, bottom tracking
                elif shp[0] == vshp[0] and not any(val in key for val in tag[:2]):
                    if 'bt' in key and 'time_bt' in data['coords']:
                        tg = '_bt'
                    else:
                        tg = ''
                    if any(key.rsplit('_')[0] in s for s in ['amp', 'corr', 'dist', 'prcnt_gd']):
                        dim0 = 'beam'
                    else:
                        dim0 = 'dir'
                    ds[key] = ds[key].rename({'dim_0': dim0,
                                              'dim_1': 'time'+tg})
                    ds[key] = ds[key].assign_coords({dim0: beam,
                                                     'time'+tg: data['coords']['time'+tg]})
                # ADCP IMU data
                elif shp[0] == vshp[0]-1:
                    if not any(val in key for val in tag):
                        tg = ''
                    else:
                        tg = [val for val in tag if val in key]
                        tg = tg[0]

                    ds[key] = ds[key].rename({'dim_0': 'dirIMU',
                                              'dim_1': 'time'+tg})
                    ds[key] = ds[key].assign_coords({'dirIMU': [1, 2, 3],
                                                     'time'+tg: data['coords']['time'+tg]})

            elif l == 3:  # 3D variables
                if 'vel' in key:
                    dim0 = 'dir'
                else:  # amp, corr, prcnt_gd, status
                    dim0 = 'beam'

                if not any(val in key for val in tag):
                    ds[key] = ds[key].rename({'dim_0': dim0,
                                              'dim_1': 'range',
                                              'dim_2': 'time'})
                    ds[key] = ds[key].assign_coords({dim0: beam,
                                                     'range': data['coords']['range'],
                                                     'time': data['coords']['time']})
                elif 'b5' in key:
                    # xarray can't handle coords of length 1
                    ds[key] = ds[key][0]
                    ds[key] = ds[key].rename({'dim_1': 'range_b5',
                                              'dim_2': 'time_b5'})
                    ds[key] = ds[key].assign_coords({'range_b5': data['coords']['range_b5'],
                                                     'time_b5': data['coords']['time_b5']})
                elif 'sl' in key:
                    ds[key] = ds[key].rename({'dim_0': dim0,
                                              'dim_1': 'range_sl',
                                              'dim_2': 'time'})
                    ds[key] = ds[key].assign_coords({'range_sl': data['coords']['range_sl'],
                                                     'time': data['coords']['time']})
                else:
                    ds = ds.drop_vars(key)
                    warnings.warn(f'Variable not included in dataset: {key}')

    # coordinate units
    r_list = [r for r in ds.coords if 'range' in r]
    for ky in r_list:
        ds[ky].attrs['units'] = 'm'

    ds.attrs = data['attrs']

    return ds



from setuptools import setup, find_packages
import os
import shutil

# Change this to True if you want to include the tests and test data
# in the distribution.
include_tests = False

try:
    # This deals with a bug where the tests aren't excluded due to not
    # rebuilding the files in this folder.
    shutil.rmtree('dolfyn.egg-info')
except OSError:
    pass

# Get the version info We do this to avoid importing __init__, which
# depends on other packages that may not yet be installed.
base_dir = os.path.abspath(os.path.dirname(__file__))
version = {}
with open(base_dir + "/dolfyn/_version.py") as fp:
    exec(fp.read(), version)


config = dict(
    name='dolfyn',
    version=version['__version__'],
    description='Doppler Ocean Library for pYthoN.',
    author='DOLfYN Developers',
    author_email='james.mcvey@pnnl.gov',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Topic :: Scientific/Engineering',
    ],
    url='http://github.com/lkilcher/dolfyn',
    packages=find_packages(exclude=['dolfyn.tests']),
    package_data={},
    install_requires=['numpy>=1.21',
                      'scipy>=1.7.0',
                      'xarray>=0.18.2',
                      'netcdf4>=1.5.7',
                      'bottleneck'],
    provides=['dolfyn'],
    scripts=['scripts/motcorrect_vector.py', 'scripts/binary2mat.py'],
)


if include_tests:
    config['packages'].append('dolfyn.tests')
    config['package_data'].update({'dolfyn.tests': ['data/*']},)

setup(**config)



from . import vector as r_vec
from . import awac as r_awac
from . import signature as r_sig
from . import rdi as r_rdi
from .base import _make_model
import numpy as np
import xarray as xr
import warnings


# The 'rotation chain'
rc = ['beam', 'inst', 'earth', 'principal']

rot_module_dict = {
    # Nortek instruments
    'vector': r_vec,
    'awac': r_awac,
    'signature': r_sig,
    'ad2cp': r_sig,

    # TRDI instruments
    'rdi': r_rdi}


def rotate2(ds, out_frame='earth', inplace=True):
    """Rotate a dataset to a new coordinate system.

    Parameters
    ----------
    ds : xr.Dataset
      The dolfyn dataset (ADV or ADCP) to rotate.
    out_frame : string {'beam', 'inst', 'earth', 'principal'}
      The coordinate system to rotate the data into.
    inplace : bool (default: True)
        When True ``ds`` is modified. When False a copy is returned.

    Returns
    -------
    ds : xarray.Dataset or None
      Returns a new rotated dataset **when ``inplace=False``**, otherwise
      returns None.

    Notes
    -----
    - This function rotates all variables in ``ds.attrs['rotate_vars']``.

    - In order to rotate to the 'principal' frame, a value should exist for 
      ``ds.attrs['principal_heading']``. The function 
      :func:`calc_principal_heading <dolfyn.calc_principal_heading>`
      is recommended for this purpose, e.g.:

          ds.attrs['principal_heading'] = dolfyn.calc_principal_heading(ds['vel'].mean(range))

      where here we are using the depth-averaged velocity to calculate
      the principal direction.

    """
    # Create and return deep copy if not writing "in place"
    if not inplace:
        ds = ds.copy(deep=True)

    csin = ds.coord_sys.lower()
    if csin == 'ship':
        csin = 'inst'

    # Returns True/False if head2inst_rotmat has been set/not-set.
    r_vec._check_inst2head_rotmat(ds)

    if out_frame == 'principal' and csin != 'earth':
        warnings.warn(
            "You are attempting to rotate into the 'principal' "
            "coordinate system, but the dataset is in the {} "
            "coordinate system. Be sure that 'principal_heading' is "
            "defined based on the earth coordinate system.".format(csin))

    rmod = None
    for ky in rot_module_dict:
        if ky in _make_model(ds):
            rmod = rot_module_dict[ky]
            break
    if rmod is None:
        raise ValueError("Rotations are not defined for "
                         "instrument '{}'.".format(_make_model(ds)))

    # Get the 'indices' of the rotation chain
    try:
        iframe_in = rc.index(csin)
    except ValueError:
        raise Exception("The coordinate system of the input "
                        "dataset, '{}', is invalid."
                        .format(ds.coord_sys))
    try:
        iframe_out = rc.index(out_frame.lower())
    except ValueError:
        raise Exception("The specified output coordinate system "
                        "is invalid, please select one of: 'beam', 'inst', "
                        "'earth', 'principal'.")

    if iframe_out == iframe_in:
        print("Data is already in the {} coordinate system".format(out_frame))

    if iframe_out > iframe_in:
        reverse = False
    else:
        reverse = True

    while ds.coord_sys.lower() != out_frame.lower():
        csin = ds.coord_sys
        if csin == 'ship':
            csin = 'inst'
        inow = rc.index(csin)
        if reverse:
            func = getattr(rmod, '_' + rc[inow - 1] + '2' + rc[inow])
        else:
            func = getattr(rmod, '_' + rc[inow] + '2' + rc[inow + 1])
        ds = func(ds, reverse=reverse)

    if not inplace:
        return ds


def calc_principal_heading(vel, tidal_mode=True):
    """Compute the principal angle of the horizontal velocity.

    Parameters
    ----------
    vel : np.ndarray (2,...,Nt), or (3,...,Nt)
      The 2D or 3D velocity array (3rd-dim is ignored in this calculation)
    tidal_mode : bool (default: True)
      If true, range is set from 0 to +/-180 degrees. If false, range is 0 to 
      360 degrees

    Returns
    -------
    p_heading : float or ndarray
      The principal heading in degrees clockwise from North.

    Notes
    -----
    When tidal_mode=True, this tool calculates the heading that is
    aligned with the bidirectional flow. It does so following these
    steps:
      1. rotates vectors with negative velocity by 180 degrees
      2. then doubles those angles to make a complete circle again
      3. computes a mean direction from this, and halves that angle
         (to undo the doubled-angles in step 2)
      4. The returned angle is forced to be between 0 and 180. So, you
         may need to add 180 to this if you want your positive
         direction to be in the western-half of the plane.

    Otherwise, this function simply computes the average direction
    using a vector method.

    """
    dt = vel[0] + vel[1] * 1j
    if tidal_mode:
        # Flip all vectors that are below the x-axis
        dt[dt.imag <= 0] *= -1
        # Now double the angle, so that angles near pi and 0 get averaged
        # together correctly:
        dt *= np.exp(1j * np.angle(dt))
        dt = np.ma.masked_invalid(dt)
        # Divide the angle by 2 to remove the doubling done on the previous
        # line.
        pang = np.angle(
            np.nanmean(dt, -1, dtype=np.complex128)) / 2
    else:
        pang = np.angle(np.nanmean(dt, -1))

    return np.round((90 - np.rad2deg(pang)), decimals=4)


def set_declination(ds, declin, inplace=True):
    """Set the magnetic declination

    Parameters
    ----------
    ds : xarray.Dataset or :class:`dolfyn.velocity.Velocity`
       The input dataset or velocity class
    declination : float
       The value of the magnetic declination in degrees (positive
       values specify that Magnetic North is clockwise from True North)
    inplace : bool (default: True)
        When True ``ds`` is modified. When False a copy is returned.

    Returns
    -------
    ds : xarray.Dataset or None
      Returns a new dataset with declination set **when
      ``inplace=False``**, otherwise returns None.

    Notes
    -----
    This function modifies the data object in the following ways:

    - If the dataset is in the *earth* reference frame at the time of
      setting declination, it will be rotated into the "*True-East*,
      *True-North*, Up" (hereafter, ETU) coordinate system

    - ``dat['orientmat']`` is modified to be an ETU to
      instrument (XYZ) rotation matrix (rather than the magnetic-ENU to
      XYZ rotation matrix). Therefore, all rotations to/from the 'earth'
      frame will now be to/from this ETU coordinate system.

    - The value of the specified declination will be stored in
      ``dat.attrs['declination']``

    - ``dat['heading']`` is adjusted for declination
      (i.e., it is relative to True North).

    - If ``dat.attrs['principal_heading']`` is set, it is
      adjusted to account for the orientation of the new 'True'
      earth coordinate system (i.e., calling set_declination on a
      data object in the principal coordinate system, then calling
      dat.rotate2('earth') will yield a data object in the new
      'True' earth coordinate system)

    """
    # Create and return deep copy if not writing "in place"
    if not inplace:
        ds = ds.copy(deep=True)

    if 'declination' in ds.attrs:
        angle = declin - ds.attrs.pop('declination')
    else:
        angle = declin
    cd = np.cos(-np.deg2rad(angle))
    sd = np.sin(-np.deg2rad(angle))

    # The ordering is funny here because orientmat is the
    # transpose of the inst->earth rotation matrix:
    Rdec = np.array([[cd, -sd, 0],
                     [sd, cd, 0],
                     [0, 0, 1]])

    if ds.coord_sys == 'earth':
        rotate2earth = True
        rotate2(ds, 'inst', inplace=True)
    else:
        rotate2earth = False

    ds['orientmat'].values = np.einsum('kj...,ij->ki...',
                                       ds['orientmat'].values,
                                       Rdec, )
    if 'heading' in ds:
        ds['heading'] += angle
    if rotate2earth:
        rotate2(ds, 'earth', inplace=True)
    if 'principal_heading' in ds.attrs:
        ds.attrs['principal_heading'] += angle

    ds.attrs['declination'] = declin
    ds.attrs['declination_in_orientmat'] = 1  # logical

    if not inplace:
        return ds


def set_inst2head_rotmat(ds, rotmat, inplace=True):
    """
    Set the instrument to head rotation matrix for the Nortek ADV if it
    hasn't already been set through a '.userdata.json' file.

    Parameters
    ----------
    ds : xarray.Dataset
        The data set to assign inst2head_rotmat
    rotmat : float
        3x3 rotation matrix
    inplace : bool (default: True)
        When True ``ds`` is modified. When False a copy is returned.

    Returns
    -------
    ds : xarray.Dataset or None
      Returns a new dataset with inst2head_rotmat set **when
      ``inplace=False``**, otherwise returns None.

    Notes
    -----
    If the data object is in earth or principal coords, it is first
    rotated to 'inst' before assigning inst2head_rotmat, it is then
    rotated back to the coordinate system in which it was input. This
    way the inst2head_rotmat gets applied correctly (in inst
    coordinate system).

    """
    # Create and return deep copy if not writing "in place"
    if not inplace:
        ds = ds.copy(deep=True)

    if not ds.inst_model.lower() == 'vector':
        raise Exception("Setting 'inst2head_rotmat' is only supported "
                        "for Nortek Vector ADVs.")
    if ds.get('inst2head_rotmat', None) is not None:
        raise Exception(
            "You are setting 'inst2head_rotmat' after it has already "
            "been set. You can only set it once.")

    csin = ds.coord_sys
    if csin not in ['inst', 'beam']:
        rotate2(ds, 'inst', inplace=True)

    ds['inst2head_rotmat'] = xr.DataArray(np.array(rotmat),
                                          dims=['x', 'x*'],
                                          coords={'x': [1, 2, 3],
                                                  'x*': [1, 2, 3]})

    ds.attrs['inst2head_rotmat_was_set'] = 1  # logical
    # Note that there is no validation that the user doesn't
    # change `ds.attrs['inst2head_rotmat']` after calling this
    # function.

    if not csin == 'beam':  # csin not 'beam', then we're in inst
        ds = r_vec._rotate_inst2head(ds)
    if csin not in ['inst', 'beam']:
        rotate2(ds, csin, inplace=True)

    if not inplace:
        return ds



"""Module containing functions to clean data
"""
import numpy as np
import warnings
from ..velocity import VelBinner
from ..tools.misc import group, slice1d_along_axis
warnings.filterwarnings('ignore', category=np.RankWarning)

sin = np.sin
cos = np.cos


def clean_fill(u, mask, npt=12, method='cubic', maxgap=6):
    """
    Interpolate over mask values in timeseries data using the specified method

    Parameters
    ----------
    u : xarray.DataArray
      The dataArray to clean.
    mask : bool
      Logical tensor of elements to "nan" out (from `spikeThresh`, `rangeLimit`,
      or `GN2002`) and replace
    npt : int
      The number of points on either side of the bad values that 
      interpolation occurs over
    method : string
      Interpolation scheme to use (linear, cubic, pchip, etc)
    limit : int
      Max number of consective nan's to interpolate across

    Returns
    -------
    da : xarray.DataArray
      The dataArray with nan's filled in

    See Also
    --------
    xarray.DataArray.interpolate_na()
    """

    # Apply mask
    u.values[..., mask] = np.nan

    # Remove bad data for 2D+ and 1D timeseries variables
    if 'dir' in u.dims:
        for i in range(u.shape[0]):
            u[i] = _interp_nan(u[i], npt, method, maxgap)
    else:
        u = _interp_nan(u, npt, method, maxgap)

    return u


def _interp_nan(da, npt, method, maxgap):
    """
    Interpolate over the points in `bad` that are True.

    Parameters
    ----------
    da : xarray.DataArray
      The field to be cleaned
    npt : int
      The number of points on either side of the gap that the fit
      occurs over

    """
    searching = True
    bds = da.isnull().values
    ntail = 0
    pos = 0
    # The index array:
    i = np.arange(len(da), dtype=np.uint32)

    while pos < len(da):
        if searching:
            # Check the point
            if bds[pos]:
                # If it's bad, mark the start
                start = max(pos - npt, 0)
                # And stop searching.
                searching = False
            pos += 1
            # Continue...
        else:
            # Another bad point?
            if bds[pos]:  # Yes
                # Reset ntail
                ntail = 0
            else:  # No
                # Add to the tail of the block.
                ntail += 1
            pos += 1

            if (ntail == npt or pos == len(da)):
                # This is the block we are interpolating over
                i_int = i[start:pos]
                da[i_int] = da[i_int].interpolate_na(dim=da.dims[-1],
                                                     method=method,
                                                     use_coordinate=True,
                                                     limit=maxgap)
                # Reset
                searching = True
                ntail = 0
    return da


def fill_nan_ensemble_mean(u, mask, fs, window):
    """
    Fill missing values with the ensemble mean.

    Parameters
    ----------
    u : xarray.DataArray (..., time)
      The dataArray to clean. Can be 1D or 2D.
    mask : bool
      Logical tensor of elements to "nan" out (from `spikeThresh`, `rangeLimit`,
      or `GN2002`) and replace
    fs : int
      Instrument sampling frequency
    window : int
      Size of window in seconds used to calculate ensemble means

    Returns
    -------
    da : xarray.DataArray
      The dataArray with nan's filled in

    Notes
    -----
    Gaps larger than the ensemble size will not get filled in.
    """

    u = u.where(~mask)
    bnr = VelBinner(n_bin=window*fs, fs=fs)

    if len(u.shape) == 1:
        var = u.values[None, :]
    else:
        var = u.values

    vel = np.empty(var.shape)
    vel_reshaped = bnr.reshape(var)
    vel_mean = np.nanmean(vel_reshaped, axis=-1)

    # If there are extra datapoints trimmed off after the last ensemble,
    # take them into account by filling in another ensemble with means
    diff = vel.shape[-1] - vel_reshaped.size // vel.shape[0]
    # diff = number of extra points
    extra_nans = vel_reshaped.shape[-1] - diff
    if diff:
        vel = np.empty((var.shape[0], var.shape[-1]+extra_nans))
        extra = var[:, -diff:]
        empty = np.empty((vel.shape[0], extra_nans))*np.nan
        extra = np.concatenate((extra, empty), axis=-1)
        vel_reshaped = np.concatenate(
            (vel_reshaped, extra[:, None, :]), axis=1)
        extra_mean = np.nanmean(extra, axis=-1)
        vel_mean = np.concatenate((vel_mean, extra_mean[:, None]), axis=-1)

    # Create a matrix the same size as the reshaped array, and mask out the
    # non-missing values. Then add the two matrices together.
    vel_mean_matrix = np.tile(vel_mean[..., None], (1, 1, bnr.n_bin))
    vel_missing = np.isnan(vel_reshaped)
    vel_mask = np.ma.masked_array(vel_mean_matrix, ~vel_missing).filled(np.nan)
    vel_filled = np.where(np.isnan(vel_reshaped), vel_mask,
                          vel_reshaped + np.nan_to_num(vel_mask))
    # "Unshape" the data
    for i in range(var.shape[0]):
        vel[i] = np.ravel(vel_filled[i], 'C')

    if diff:  # Trim off the extra means
        u.values = np.squeeze(vel[:, :-extra_nans])
    else:
        u.values = np.squeeze(vel)

    return u


def spike_thresh(u, thresh=10):
    """
    Returns a logical vector where a spike in `u` of magnitude greater than
    `thresh` occurs. Both 'Negative' and 'positive' spikes are found.

    Parameters
    ----------
    u : xarray.DataArray
      The timeseries data to clean.
    thresh : int
       Magnitude of velocity spike, must be positive.

    Returns
    -------
    mask : |np.ndarray|
      Logical vector with spikes labeled as 'True'
    """

    du = np.diff(u.values, prepend=0)
    mask = (du > thresh) + (du < -thresh)

    return mask


def range_limit(u, range=[-5, 5]):
    """
    Returns a logical vector that is True where the values of `u` are 
    outside of `range`.

    Parameters
    ----------
    u : xarray.DataArray
      The timeseries data to clean.
    range : list
       Min and max magnitudes beyond which are masked

    Returns
    -------
    mask : |np.ndarray|
      Logical vector with spikes labeled as 'True'
    """

    return ~((range[0] < u.values) & (u.values < range[1]))


def _calcab(al, Lu_std_u, Lu_std_d2u):
    """Solve equations 10 and 11 of Goring+Nikora2002
    """
    return tuple(np.linalg.solve(
        np.array([[cos(al) ** 2, sin(al) ** 2],
                  [sin(al) ** 2, cos(al) ** 2]]),
        np.array([(Lu_std_u) ** 2, (Lu_std_d2u) ** 2])))


def _phaseSpaceThresh(u):
    if u.ndim == 1:
        u = u[:, None]
    u = np.array(u)
    Lu = (2 * np.log(u.shape[0])) ** 0.5
    u = u - u.mean(0)
    du = np.zeros_like(u)
    d2u = np.zeros_like(u)
    # Take the centered difference.
    du[1:-1] = (u[2:] - u[:-2]) / 2
    # And again.
    d2u[2:-2] = (du[1:-1][2:] - du[1:-1][:-2]) / 2
    p = (u ** 2 + du ** 2 + d2u ** 2)
    std_u = np.std(u, axis=0)
    std_du = np.std(du, axis=0)
    std_d2u = np.std(d2u, axis=0)
    alpha = np.arctan2(np.sum(u * d2u, axis=0), np.sum(u ** 2, axis=0))
    a = np.empty_like(alpha)
    b = np.empty_like(alpha)
    with warnings.catch_warnings() as w:
        warnings.filterwarnings(
            'ignore', category=RuntimeWarning, message='invalid value encountered in ')
        for idx, al in enumerate(alpha):
            a[idx], b[idx] = _calcab(al, Lu * std_u[idx], Lu * std_d2u[idx])
        theta = np.arctan2(du, u)
        phi = np.arctan2((du ** 2 + u ** 2) ** 0.5, d2u)
        pe = (((sin(phi) * cos(theta) * cos(alpha) +
                cos(phi) * sin(alpha)) ** 2) / a +
              ((sin(phi) * cos(theta) * sin(alpha) -
                cos(phi) * cos(alpha)) ** 2) / b +
              ((sin(phi) * sin(theta)) ** 2) / (Lu * std_du) ** 2) ** -1
    pe[:, np.isnan(pe[0, :])] = 0
    return (p > pe).flatten('F')


def GN2002(u, npt=5000):
    """
    The Goring & Nikora 2002 'despiking' method, with Wahl2003 correction.
    Returns a logical vector that is true where spikes are identified.

    Parameters
    ----------
    u : xarray.DataArray
      The velocity array (1D or 3D) to clean.
    npt : int
      The number of points over which to perform the method.

    Returns
    -------
    mask : |np.ndarray|
      Logical vector with spikes labeled as 'True'
    """

    if not isinstance(u, np.ndarray):
        return GN2002(u.values, npt=npt)

    if u.ndim > 1:
        mask = np.zeros(u.shape, dtype='bool')
        for slc in slice1d_along_axis(u.shape, -1):
            mask[slc] = GN2002(u[slc], npt=npt)
        return mask

    mask = np.zeros(len(u), dtype='bool')

    # Find large bad segments (>npt/10):
    # group returns a vector of slice objects.
    bad_segs = group(np.isnan(u), min_length=int(npt//10))
    if bad_segs.size > 2:
        # Break them up into separate regions:
        sp = 0
        ep = len(u)

        # Skip start and end bad_segs:
        if bad_segs[0].start == sp:
            sp = bad_segs[0].stop
            bad_segs = bad_segs[1:]
        if bad_segs[-1].stop == ep:
            ep = bad_segs[-1].start
            bad_segs = bad_segs[:-1]

        for ind in range(len(bad_segs)):
            bs = bad_segs[ind]  # bs is a slice object.
            # Clean the good region:
            mask[sp:bs.start] = GN2002(u[sp:bs.start], npt=npt)
            sp = bs.stop
        # Clean the last good region.
        mask[sp:ep] = GN2002(u[sp:ep], npt=npt)
        return mask

    c = 0
    ntot = len(u)
    nbins = int(ntot // npt)
    mask_last = np.zeros_like(mask) + np.inf
    mask[0] = True  # make sure we start.
    while mask.any():
        mask[:nbins * npt] = _phaseSpaceThresh(
            np.array(np.reshape(u[:(nbins * npt)], (npt, nbins), order='F')))
        mask[-npt:] = _phaseSpaceThresh(u[-npt:])
        c += 1
        if c >= 100:
            raise Exception('GN2002 loop-limit exceeded.')
        if mask.sum() >= mask_last.sum():
            break
        mask_last = mask.copy()
    return mask



"""
This is the Doppler Oceanography Library for pYthoN (DOLfYN). It is
designed to read and work with oceanographyic velocity measurements
from Acoustic Doppler Current Profilers (ADPs/ADCPs) and Acoustic Doppler
Velocimeters (ADVs). It is a high-level object-oriented library
composed of a set of data-object classes (types) that contain data
from a particular measurement instrument, and a collection of
functions that manipulate those data objects to accomplish data
processing and data analysis tasks.

"""

from ._version import __version__
from .io.api import read, read_example, save, load, save_mat, load_mat
from .rotate.api import rotate2, calc_principal_heading, set_declination, set_inst2head_rotmat
from .rotate.base import euler2orient, orient2euler, quaternion2orient
from .velocity import VelBinner
from dolfyn import adv
from dolfyn import adp
from dolfyn import time
from dolfyn import io
from dolfyn import rotate
from dolfyn import tools



import numpy as np
from scipy.signal import medfilt2d, convolve2d


def _nans(*args, **kwargs):
    out = np.empty(*args, **kwargs)
    if np.issubdtype(out.flatten()[0], np.integer):
        out[:] = 0
        return out
    else:
        out[:] = np.nan
        return out


def _nans_like(*args, **kwargs):
    out = np.empty_like(*args, **kwargs)
    out[:] = np.nan
    return out


def _find(arr):
    return np.nonzero(np.ravel(arr))[0]


def detrend(arr, axis=-1, in_place=False):
    """Remove a linear trend from arr.

    Parameters
    ----------
    arr : array_like
       The array from which to remove a linear trend.
    axis : int
       The axis along which to operate.

    Notes
    -----
    This method is copied from the matplotlib.mlab library, but
    implements the covariance calcs explicitly for added speed.

    This works much faster than mpl.mlab.detrend for multi-dimensional
    arrays, and is also faster than linalg.lstsq methods.
    """
    arr = np.asarray(arr)
    if not in_place:
        arr = arr.copy()
    sz = np.ones(arr.ndim, dtype=int)
    sz[axis] = arr.shape[axis]
    x = np.arange(sz[axis], dtype=np.float_).reshape(sz)
    x -= np.nanmean(x, axis=axis, keepdims=True)
    arr -= np.nanmean(arr, axis=axis, keepdims=True)
    b = np.nanmean((x * arr), axis=axis, keepdims=True) / \
        np.nanmean((x ** 2), axis=axis, keepdims=True)
    arr -= b * x
    return arr


def group(bl, min_length=0):
    """Find continuous segments in a boolean array.

    Parameters
    ----------
    bl : |np.ndarray| (dtype='bool')
      The input boolean array.
    min_length : int (optional)
      Specifies the minimum number of continuous points to consider a
      `group` (i.e. that will be returned).

    Returns
    -------
    out : np.ndarray(slices,)
      a vector of slice objects, which indicate the continuous
      sections where `bl` is True.

    Notes
    -----
    This function has funny behavior for single points.  It will
    return the same two indices for the beginning and end.

    """
    if not any(bl):
        return np.empty(0)
    vl = np.diff(bl.astype('int'))
    ups = np.nonzero(vl == 1)[0] + 1
    dns = np.nonzero(vl == -1)[0] + 1
    if bl[0]:
        if len(ups) == 0:
            ups = np.array([0])
        else:
            ups = np.concatenate((np.array([0]), [len(ups)]))
    if bl[-1]:
        if len(dns) == 0:
            dns = np.array([len(bl)])
        else:
            dns = np.concatenate((dns, [len(bl)]))
    out = np.empty(len(dns), dtype='O')
    idx = 0
    for u, d in zip(ups, dns):
        if d - u < min_length:
            continue
        out[idx] = slice(u, d)
        idx += 1
    return out[:idx]


def slice1d_along_axis(arr_shape, axis=0):
    """
    Return an iterator object for looping over 1-D slices, along ``axis``, of
    an array of shape arr_shape.

    Parameters
    ----------
    arr_shape : tuple,list
      Shape of the array over which the slices will be made.
    axis : integer
      Axis along which `arr` is sliced.

    Returns
    -------
    Iterator : object
      The iterator object returns slice objects which slices arrays of
      shape arr_shape into 1-D arrays.

    Examples
    --------
    >> out=np.empty(replace(arr.shape,0,1))
    >> for slc in slice1d_along_axis(arr.shape,axis=0):
    >>     out[slc]=my_1d_function(arr[slc])

    """
    nd = len(arr_shape)
    if axis < 0:
        axis += nd
    ind = [0] * (nd - 1)
    i = np.zeros(nd, 'O')
    indlist = list(range(nd))
    indlist.remove(axis)
    i[axis] = slice(None)
    itr_dims = np.asarray(arr_shape).take(indlist)
    Ntot = np.product(itr_dims)
    i.put(indlist, ind)
    k = 0
    while k < Ntot:
        # increment the index
        n = -1
        while (ind[n] >= itr_dims[n]) and (n > (1 - nd)):
            ind[n - 1] += 1
            ind[n] = 0
            n -= 1
        i.put(indlist, ind)
        yield tuple(i)
        ind[-1] += 1
        k += 1


def fillgaps(a, maxgap=np.inf, dim=0, extrapFlg=False):
    """Linearly fill NaN value in an array.

    Parameters
    ----------
    a : |np.ndarray|
      The array to be filled.
    maxgap : |np.ndarray| (optional: inf)
      The maximum gap to fill.
    dim : int (optional: 0)
      The dimension to operate along.
    extrapFlg : bool (optional: False)
      Whether to extrapolate if NaNs are found at the ends of the
      array.

    See Also
    --------
    dolfyn.tools.misc.interpgaps : Linearly interpolates in time.

    Notes
    -----
    This function interpolates assuming spacing/timestep between
    successive points is constant. If the spacing is not constant, use
    interpgaps.

    """

    # If this is a multi-dimensional array, operate along axis dim.
    if a.ndim > 1:
        for inds in slice1d_along_axis(a.shape, dim):
            fillgaps(a[inds], maxgap, 0, extrapFlg)
        return

    a = np.asarray(a)
    nd = a.ndim
    if dim < 0:
        dim += nd
    if (dim >= nd):
        raise ValueError("dim must be less than a.ndim; dim=%d, rank=%d."
                         % (dim, nd))
    ind = [0] * (nd - 1)
    i = np.zeros(nd, 'O')
    indlist = list(range(nd))
    indlist.remove(dim)
    i[dim] = slice(None, None)
    i.put(indlist, ind)

    gd = np.nonzero(~np.isnan(a))[0]

    # Here we extrapolate the ends, if necessary:
    if extrapFlg and gd.__len__() > 0:
        if gd[0] != 0 and gd[0] <= maxgap:
            a[:gd[0]] = a[gd[0]]
        if gd[-1] != a.__len__() and (a.__len__() - (gd[-1] + 1)) <= maxgap:
            a[gd[-1]:] = a[gd[-1]]

    # Here is the main loop
    if gd.__len__() > 1:
        inds = np.nonzero((1 < np.diff(gd)) & (np.diff(gd) <= maxgap + 1))[0]
        for i2 in range(0, inds.__len__()):
            ii = list(range(gd[inds[i2]] + 1, gd[inds[i2] + 1]))
            a[ii] = (np.diff(a[gd[[inds[i2], inds[i2] + 1]]]) *
                     (np.arange(0, ii.__len__()) + 1) /
                     (ii.__len__() + 1) + a[gd[inds[i2]]]).astype(a.dtype)

    return a


def interpgaps(a, t, maxgap=np.inf, dim=0, extrapFlg=False):
    """
    Fill gaps (NaN values) in ``a`` by linear interpolation along
    dimension ``dim`` with the point spacing specified in ``t``.

    Parameters
    ----------
    a : |np.ndarray|
      The array containing NaN values to be filled.
    t : |np.ndarray| (len(t) == a.shape[dim])
      Independent variable of the points in ``a``, e.g. timestep
    maxgap : |np.ndarray| (optional: inf)
      The maximum gap to fill.
    dim : int (optional: 0)
      The dimension to operate along.
    extrapFlg : bool (optional: False)
      Whether to extrapolate if NaNs are found at the ends of the
      array.

    See Also
    --------
    dolfyn.tools.misc.fillgaps : Linearly interpolates in array-index space.

    """

    # If this is a multi-dimensional array, operate along dim dim.
    if a.ndim > 1:
        for inds in slice1d_along_axis(a.shape, dim):
            interpgaps(a[inds], t, maxgap, 0, extrapFlg)
        return

    gd = _find(~np.isnan(a))

    # Here we extrapolate the ends, if necessary:
    if extrapFlg and gd.__len__() > 0:
        if gd[0] != 0 and gd[0] <= maxgap:
            a[:gd[0]] = a[gd[0]]
        if gd[-1] != a.__len__() and (a.__len__() - (gd[-1] + 1)) <= maxgap:
            a[gd[-1]:] = a[gd[-1]]

    # Here is the main loop
    if gd.__len__() > 1:
        inds = _find((1 < np.diff(gd)) &
                     (np.diff(gd) <= maxgap + 1))
        for i2 in range(0, inds.__len__()):
            ii = np.arange(gd[inds[i2]] + 1, gd[inds[i2] + 1])
            ti = (t[ii] - t[gd[inds[i2]]]) / np.diff(t[[gd[inds[i2]],
                                                        gd[inds[i2] + 1]]])
            a[ii] = (np.diff(a[gd[[inds[i2], inds[i2] + 1]]]) * ti +
                     a[gd[inds[i2]]]).astype(a.dtype)

    return a


def medfiltnan(a, kernel, thresh=0):
    """
    Do a running median filter of the data. Regions where more than 
    ``thresh`` fraction of the points are NaN are set to NaN.

    Parameters
    ----------
    a : |np.ndarray|
      2D array containing data to be filtered.
    kernel_size : |np.ndarray| or list, optional
      A scalar or a list of length 2, giving the size of the median 
      filter window in each dimension. Elements of kernel_size should 
      be odd. If kernel_size is a scalar, then this scalar is used as 
      the size in each dimension.
    thresh : int
      Maximum gap in *a* to filter over

    Returns
    -------
    out : |np.ndarray|
      2D array of same size containing filtered data

    See Also
    --------
    scipy.signal.medfilt2d

    """
    flag_1D = False
    if a.ndim == 1:
        a = a[None, :]
        flag_1D = True
    try:
        len(kernel)
    except:
        kernel = [1, kernel]
    out = medfilt2d(a, kernel)
    if thresh > 0:
        out[convolve2d(np.isnan(a),
                       np.ones(kernel) / np.prod(kernel),
                       'same') > thresh] = np.NaN
    if flag_1D:
        return out[0]
    return out


def convert_degrees(deg, tidal_mode=True):
    """
    Converts between the 'cartesian angle' (counter-clockwise from East) and
    the 'polar angle' in (degrees clockwise from North)

    Parameters
    ----------
    deg: float or array-like
      Number or array in 'degrees CCW from East' or 'degrees CW from North'
    tidal_mode : bool (default: True)
      If true, range is set from 0 to +/-180 degrees. If false, range is 0 to 
      360 degrees

    Returns
    -------
    out : float or array-like
      Input data transformed to 'degrees CW from North' or 
      'degrees CCW from East', respectively (based on `deg`)

    Notes
    -----
    The same algorithm is used to convert back and forth between 'CCW from E' 
    and 'CW from N'

    """
    out = -(deg - 90) % 360
    if tidal_mode:
        out[out > 180] -= 360
    return out



import numpy as np
from .misc import detrend
fft = np.fft.fft


def psd_freq(nfft, fs, full=False):
    """
    Compute the frequency for vector for a `nfft` and `fs`.

    Parameters
    ----------
    fs : float
      The sampling frequency (e.g. samples/sec)
    nfft : int
      The number of samples in a window.
    full : bool (default: False)
      Whether to return half frequencies (positive), or the full frequencies.

    Returns
    -------
    freq : |np.ndarray|
      The frequency vector, in same units as 'fs'

    """
    fs = np.float64(fs)
    f = np.fft.fftfreq(int(nfft), 1 / fs)
    if full:
        return f
    else:
        return np.abs(f[1:int(nfft / 2. + 1)])


def _getwindow(window, nfft):
    if window == 'hann':
        window = np.hanning(nfft)
    elif window == 'hamm':
        window = np.hamming(nfft)
    elif window is None or window == 1:
        window = np.ones(nfft)
    return window


def stepsize(l, nfft, nens=None, step=None):
    """
    Calculates the fft-step size for a length *l* array.

    If nens is None, the step size is chosen to maximize data use,
    minimize nens and have a minimum of 50% overlap.

    If nens is specified, the step-size is computed directly.

    Parameters
    ----------
    l       : The length of the array.
    nfft    : The number of points in the fft.
    nens : The number of nens to perform (default compute this).

    Returns
    -------
    step    : The step size.
    nens    : The number of ensemble ffts to average together.
    nfft    : The number of points in the fft (set to l if nfft>l).

    """
    if l < nfft:
        nfft = l
    if nens is None and step is None:
        if l == nfft:
            return 0, 1, int(nfft)
        nens = int(2. * l / nfft)
        return int((l - nfft) / (nens - 1)), nens, int(nfft)
    elif nens is None:
        return int(step), int((l - nfft) / step + 1), int(nfft)
    else:
        if nens == 1:
            return 0, 1, int(nfft)
        return int((l - nfft) / (nens - 1)), int(nens), int(nfft)


def coherence(a, b, nfft, window='hann', debias=True, noise=(0, 0)):
    """
    Computes the magnitude-squared coherence of `a` and `b`.

    Parameters
    ----------
    a : |np.ndarray|
      The first array over which to compute coherence.
    b : |np.ndarray|
      The second array over which to compute coherence.
    nfft : int
      The number of points to use in the fft.
    window : string, np.ndarray (default 'hann')
      The window to use for ffts.
    debias : bool (default: True)
      Specify whether to debias the signal according to Benignus1969.
    noise : tuple(2), or float
      The `noise` keyword may be used to specify the signals'
      noise levels (std of noise in a,b). If `noise` is a two
      element tuple or list, the first and second elements specify
      the noise levels of a and b, respectively.
      default: noise=(0,0)

    Returns
    -------
    out : |np.ndarray|
      Coherence between `a` and `b`

    Notes
    -----
    Coherence is defined as:

    .. math::

      C_{ab} = \\frac{|S_{ab}|^2}{S_{aa} * S_{bb}}

    Here :math:`S_{ab}`, :math:`S_{aa}` and :math:`S_{bb}` are the cross,
    and auto spectral densities of the signal `a` and `b`.

    """
    l = [len(a), len(b)]
    cross = cpsd_quasisync
    if l[0] == l[1]:
        cross = cpsd
    elif l[0] > l[1]:
        a, b = b, a
        l = l[::-1]
    step1, nens, nfft = stepsize(l[0], nfft)
    step2, nens, nfft = stepsize(l[1], nfft, nens=nens)
    if noise.__class__ not in [list, tuple, np.ndarray]:
        noise = [noise, noise]
    elif len(noise) == 1:
        noise = [noise[0], noise[0]]
    if nens <= 2:
        raise Exception("Coherence must be computed from a set of ensembles.")
    # fs=1 is ok because it comes out in the normalization.  (noise
    # normalization depends on this)
    out = ((np.abs(cross(a, b, nfft, 1, window=window)) ** 2) /
           ((psd(a, nfft, 1, window=window, step=step1) - noise[0] ** 2 / np.pi) *
            (psd(b, nfft, 1, window=window, step=step2) - noise[1] ** 2 / np.pi))
           )
    if debias:
        # This is from Benignus1969, it seems to work (make data with different
        # nens (nfft) agree).
        return out * (1 + 1. / nens) - 1. / nens
    return out


def cpsd_quasisync(a, b, nfft, fs, window='hann'):
    """
    Compute the cross power spectral density (CPSD) of the signals `a` and `b`.

    Parameters
    ----------
    a : |np.ndarray|
      The first signal.
    b : |np.ndarray|
      The second signal.
    nfft : int
      The number of points in the fft.
    fs : float
      The sample rate (e.g. sample/second).
    window : {None, 1, 'hann', |np.ndarray|}
      The window to use (default: 'hann'). Valid entries are:
      - None,1               : uses a 'boxcar' or ones window.
      - 'hann'               : hanning window.
      - a length(nfft) array : use this as the window directly.

    Returns
    -------
    cpsd : |np.ndarray|
      The cross-spectral density of `a` and `b`.

    See Also
    ---------
    :func:`psd`,
    :func:`coherence`,
    :func:`cpsd`,
    numpy.fft

    Notes
    -----
    `a` and `b` do not need to be 'tightly' synchronized, and can even
    be different lengths, but the first- and last-index of both series
    should be synchronized (to whatever degree you want unbiased
    phases).

    This performs:

    .. math::

        fft(a)*conj(fft(b))

    Note that this is consistent with :func:`numpy.correlate`.

    It detrends the data and uses a minimum of 50% overlap for the
    shorter of `a` and `b`. For the longer, the overlap depends on the
    difference in size.  1-(l_short/l_long) data will be underutilized
    (where l_short and l_long are the length of the shorter and longer
    series, respectively).

    The units of the spectra is the product of the units of `a` and
    `b`, divided by the units of fs.

    """
    if np.iscomplexobj(a) or np.iscomplexobj(b):
        raise Exception("Velocity cannot be complex")
    l = [len(a), len(b)]
    if l[0] == l[1]:
        return cpsd(a, b, nfft, fs, window=window)
    elif l[0] > l[1]:
        a, b = b, a
        l = l[::-1]
    step = [0, 0]
    step[0], nens, nfft = stepsize(l[0], nfft)
    step[1], nens, nfft = stepsize(l[1], nfft, nens=nens)
    fs = np.float64(fs)
    window = _getwindow(window, nfft)
    fft_inds = slice(1, int(nfft / 2. + 1))
    wght = 2. / (window ** 2).sum()
    pwr = fft(detrend(a[0:nfft]) * window)[fft_inds] * \
        np.conj(fft(detrend(b[0:nfft]) * window)[fft_inds])
    if nens - 1:
        for i1, i2 in zip(range(step[0], l[0] - nfft + 1, step[0]),
                          range(step[1], l[1] - nfft + 1, step[1])):
            pwr += fft(detrend(a[i1:(i1 + nfft)]) * window)[fft_inds] * \
                np.conj(fft(detrend(b[i2:(i2 + nfft)]) * window)[fft_inds])
    pwr *= wght / nens / fs
    return pwr


def cpsd(a, b, nfft, fs, window='hann', step=None):
    """
    Compute the cross power spectral density (CPSD) of the signals `a` and `b`.

    Parameters
    ----------
    a : |np.ndarray|
      The first signal.
    b : |np.ndarray|
      The second signal.
    nfft : int
      The number of points in the fft.
    fs : float
      The sample rate (e.g. sample/second).
    window : {None, 1, 'hann', |np.ndarray|}
      The window to use (default: 'hann'). Valid entries are:
      - None,1               : uses a 'boxcar' or ones window.
      - 'hann'               : hanning window.
      - a length(nfft) array : use this as the window directly.
    step : int
      Use this to specify the overlap.  For example:
      - step : nfft/2 specifies a 50% overlap.
      - step : nfft specifies no overlap.
      - step=2*nfft means that half the data will be skipped.
      By default, `step` is calculated to maximize data use, have
      at least 50% overlap and minimize the number of ensembles.

    Returns
    -------
    cpsd : |np.ndarray|
      The cross-spectral density of `a` and `b`.

    See also
    --------
    :func:`psd`
    :func:`coherence`
    `numpy.fft`

    Notes
    -----
    cpsd removes a linear trend from the signals.

    The two signals should be the same length, and should both be real.

    This performs:

    .. math::

        fft(a)*conj(fft(b))

    This implementation is consistent with the numpy.correlate
    definition of correlation.  (The conjugate of D.B. Chelton's
    definition of correlation.)

    The units of the spectra is the product of the units of `a` and
    `b`, divided by the units of fs.

    """
    if np.iscomplexobj(a) or np.iscomplexobj(b):
        raise Exception("Velocity cannot be complex")
    auto_psd = False
    if a is b:
        auto_psd = True
    l = len(a)
    step, nens, nfft = stepsize(l, nfft, step=step)
    fs = np.float64(fs)
    window = _getwindow(window, nfft)
    fft_inds = slice(1, int(nfft / 2. + 1))
    wght = 2. / (window ** 2).sum()
    s1 = fft(detrend(a[0:nfft]) * window)[fft_inds]
    if auto_psd:
        pwr = np.abs(s1) ** 2
    else:
        pwr = s1 * np.conj(fft(detrend(b[0:nfft]) * window)[fft_inds])
    if nens - 1:
        for i in range(step, l - nfft + 1, step):
            s1 = fft(detrend(a[i:(i + nfft)]) * window)[fft_inds]
            if auto_psd:
                pwr += np.abs(s1) ** 2
            else:
                pwr += s1 * \
                    np.conj(fft(detrend(b[i:(i + nfft)]) * window)[fft_inds])
    pwr *= wght / nens / fs
    return pwr


def psd(a, nfft, fs, window='hann', step=None):
    """
    Compute the power spectral density (PSD).

    This function computes the one-dimensional `n`-point PSD.

    The units of the spectra is the product of the units of `a` and
    `b`, divided by the units of fs.

    Parameters
    ----------
    a : |np.ndarray|
      The first signal, as a 1D vector
    nfft : int
      The number of points in the fft.
    fs : float
      The sample rate (e.g. sample/second).
    window : {None, 1, 'hann', |np.ndarray|}
      The window to use (default: 'hann'). Valid entries are:
      - None,1               : uses a 'boxcar' or ones window.
      - 'hann'               : hanning window.
      - a length(nfft) array : use this as the window directly.
    step : int
      Use this to specify the overlap.  For example:
      - step : nfft/2 specifies a 50% overlap.
      - step : nfft specifies no overlap.
      - step=2*nfft means that half the data will be skipped.
      By default, `step` is calculated to maximize data use, have
      at least 50% overlap and minimize the number of ensembles.

    Returns
    -------
    psd : |np.ndarray|
      The power spectral density of `a` and `b`.

    Notes
    -----
    Credit: This function's line of code was copied from JN's fast_psd.m 
    routine.

    See Also
    --------
    :func:`cpsd`
    :func:`coherence`
    `numpy.fft`

    """
    return np.abs(cpsd(a, a, nfft, fs, window=window, step=step))


def phase_angle(a, b, nfft, window='hann', step=None):
    """
    Compute the phase difference between signals `a` and `b`. This is the
    complimentary function to coherence and cpsd.

    Positive angles means that `b` leads `a`, i.e. this does,
    essentially:

          angle(b) - angle(a)

    This function computes one-dimensional `n`-point PSD.

    The angles are output as magnitude = 1 complex numbers (to
    simplify averaging). Therefore, use `numpy.angle` to actually
    output the angle.


    Parameters
    ----------
    a      : 1d-array_like, the signal. Currently only supports vectors.
    nfft   : The number of points in the fft.
    window : The window to use (default: 'hann'). Valid entries are:
                 None,1               : uses a 'boxcar' or ones window.
                 'hann'               : hanning window.
                 a length(nfft) array : use this as the window directly.
    step   : Use this to specify the overlap.  For example:
             -  step=nfft/2 specifies a 50% overlap.
             -  step=nfft specifies no overlap.
             -  step=2*nfft means that half the data will be skipped.

             By default, `step` is calculated to maximize data use, have
             at least 50% overlap and minimize the number of ensembles.

    Returns
    -------
    ang    : complex |np.ndarray| (unit magnitude values)

    See Also
    --------
    `numpy.fft`
    :func:`coherence`
    :func:`cpsd`

    """
    window = _getwindow(window, nfft)
    fft_inds = slice(1, int(nfft / 2. + 1))
    s1 = fft(detrend(a[0:nfft]) * window)[fft_inds]
    s2 = fft(detrend(b[0:nfft]) * window)[fft_inds]
    s1 /= np.abs(s1)
    s2 /= np.abs(s2)
    ang = s2 / s1
    l = len(a)
    step, nens, nfft = stepsize(l, nfft, step=step)
    if nens - 1:
        for i in range(step, l - nfft + 1, step):
            s1 = fft(detrend(a[i:(i + nfft)]) * window)[fft_inds]
            s1 /= np.abs(s1)
            s2 = fft(detrend(a[i:(i + nfft)]) * window)[fft_inds]
            s2 /= np.abs(s2)
            ang += s2 / s1
    ang /= nens
    return ang



from datetime import datetime, timedelta, timezone
import numpy as np
from .tools.misc import fillgaps


def _fullyear(year):
    if year > 100:
        return year
    year += 1900 + 100 * (year < 90)
    return year


def epoch2dt64(ep_time):
    """
    Convert from epoch time (seconds since 1/1/1970 00:00:00) to 
    numpy.datetime64 array

    Parameters
    ----------
    ep_time : xarray.DataArray
        Time coordinate data-array or single time element

    Returns
    -------
    time : numpy.datetime64
        The converted datetime64 array

    """
    # assumes t0=1970-01-01 00:00:00
    out = np.array(ep_time.astype('int')).astype('datetime64[s]')
    out = out + ((ep_time % 1) * 1e9).astype('timedelta64[ns]')
    return out


def dt642epoch(dt64):
    """
    Convert numpy.datetime64 array to epoch time 
    (seconds since 1/1/1970 00:00:00)

    Parameters
    ----------
    dt64 : numpy.datetime64
        Single or array of datetime64 object(s)

    Returns
    -------
    time : float
        Epoch time (seconds since 1/1/1970 00:00:00)

    """
    return dt64.astype('datetime64[ns]').astype('float') / 1e9


def date2dt64(dt):
    """
    Convert numpy.datetime64 array to list of datetime objects

    Parameters
    ----------
    time : datetime.datetime
        The converted datetime object

    Returns
    -------
    dt64 : numpy.datetime64
        Single or array of datetime64 object(s)

    """
    return np.array(dt).astype('datetime64[ns]')


def dt642date(dt64):
    """
    Convert numpy.datetime64 array to list of datetime objects

    Parameters
    ----------
    dt64 : numpy.datetime64
        Single or array of datetime64 object(s)

    Returns
    -------
    time : datetime.datetime
        The converted datetime object

    """
    return epoch2date(dt642epoch(dt64))


def epoch2date(ep_time, offset_hr=0, to_str=False):
    """
    Convert from epoch time (seconds since 1/1/1970 00:00:00) to a list 
    of datetime objects

    Parameters
    ----------
    ep_time : xarray.DataArray
        Time coordinate data-array or single time element
    offset_hr : int
        Number of hours to offset time by (e.g. UTC -7 hours = PDT)
    to_str : logical
        Converts datetime object to a readable string

    Returns
    -------
    time : datetime.datetime
        The converted datetime object or list(strings) 

    Notes
    -----
    The specific time instance is set during deployment, usually sync'd to the
    deployment computer. The time seen by |dlfn| is in the timezone of the 
    deployment computer, which is unknown to |dlfn|.

    """
    try:
        ep_time = ep_time.values
    except AttributeError:
        pass

    if isinstance(ep_time, (np.ndarray)) and ep_time.ndim == 0:
        ep_time = [ep_time.item()]
    elif not isinstance(ep_time, (np.ndarray, list)):
        ep_time = [ep_time]

    ######### IMPORTANT #########
    # Note the use of `utcfromtimestamp` here, rather than `fromtimestamp`
    # This is CRITICAL! See the difference between those functions here:
    #    https://docs.python.org/3/library/datetime.html#datetime.datetime.fromtimestamp
    # Long story short: `fromtimestamp` used system-specific timezone
    # info to calculate the datetime object, but returns a
    # timezone-agnostic object.
    if offset_hr != 0:
        delta = timedelta(hours=offset_hr)
        time = [datetime.utcfromtimestamp(t) + delta for t in ep_time]
    else:
        time = [datetime.utcfromtimestamp(t) for t in ep_time]

    if to_str:
        time = date2str(time)

    return time


def date2str(dt, format_str=None):
    """
    Convert list of datetime objects to legible strings

    Parameters
    ----------
    dt : datetime.datetime
        Single or list of datetime object(s)
    format_str : string
        Timestamp string formatting, default: '%Y-%m-%d %H:%M:%S.%f'. 
        See datetime.strftime documentation for timestamp string formatting

    Returns
    -------
    time : string
        Converted timestamps

    """
    if format_str is None:
        format_str = '%Y-%m-%d %H:%M:%S.%f'

    if not isinstance(dt, list):
        dt = [dt]

    return [t.strftime(format_str) for t in dt]


def date2epoch(dt):
    """
    Convert list of datetime objects to epoch time

    Parameters
    ----------
    dt : datetime.datetime
        Single or list of datetime object(s)

    Returns
    -------
    time : float
        Datetime converted to epoch time (seconds since 1/1/1970 00:00:00)

    """
    if not isinstance(dt, list):
        dt = [dt]

    return [t.replace(tzinfo=timezone.utc).timestamp() for t in dt]


def date2matlab(dt):
    """
    Convert list of datetime objects to MATLAB datenum

    Parameters
    ----------
    dt : datetime.datetime
        List of datetime objects

    Returns
    -------
    time : float
        List of timestamps in MATLAB datnum format

    """
    time = list()
    for i in range(len(dt)):
        mdn = dt[i] + timedelta(days=366)
        frac_seconds = (dt[i]-datetime(dt[i].year, dt[i].month,
                        dt[i].day, 0, 0, 0)).seconds / (24*60*60)
        frac_microseconds = dt[i].microsecond / (24*60*60*1000000)
        time.append(mdn.toordinal() + frac_seconds + frac_microseconds)

    return time


def matlab2date(matlab_dn):
    """
    Convert MATLAB datenum to list of datetime objects

    Parameters
    ----------
    matlab_dn : float
        List of timestamps in MATLAB datnum format

    Returns
    -------
    dt : datetime.datetime
        List of datetime objects

    """
    time = list()
    for i in range(len(matlab_dn)):
        day = datetime.fromordinal(int(matlab_dn[i]))
        dayfrac = timedelta(days=matlab_dn[i] % 1) - timedelta(days=366)
        time.append(day + dayfrac)

        # Datenum is precise down to 100 microseconds - add difference to round
        us = int(round(time[i].microsecond/100, 0))*100
        time[i] = time[i].replace(microsecond=time[i].microsecond) + \
            timedelta(microseconds=us-time[i].microsecond)

    return time


def _fill_time_gaps(epoch, sample_rate_hz):
    """Fill gaps (NaN values) in the timeseries by simple linear
    interpolation.  The ends are extrapolated by stepping
    forward/backward by 1/sample_rate_hz.
    """
    # epoch is seconds since 1970
    dt = 1. / sample_rate_hz
    epoch = fillgaps(epoch)
    if np.isnan(epoch[0]):
        i0 = np.nonzero(~np.isnan(epoch))[0][0]
        delta = np.arange(-i0, 0, 1) * dt
        epoch[:i0] = epoch[i0] + delta
    if np.isnan(epoch[-1]):
        # Search backward through the array to get the 'negative index'
        ie = -np.nonzero(~np.isnan(epoch[::-1]))[0][0] - 1
        delta = np.arange(1, -ie, 1) * dt
        epoch[(ie + 1):] = epoch[ie] + delta

    return epoch
